package stockMarketApplet.GUI;

import java.awt.*;
import java.util.*;
import java.applet.*;
import javax.swing.*;
import java.awt.event.*;

import stockMarketApplet.*;
import stockMarketApplet.GUI.*;
import stockMarketApplet.util.*;
import stockMarketApplet.market.*;
import stockMarketApplet.market.io.*;

// This class opens a window with several scrollbars and an optional List.
// When the OK button is pressed, the changed values are directly set.
public class OptFrame extends Frame {

  private static final int CHEN = 0;
  private static final int EXCESS = 1;
  private TradersParametersRecord tradersParamRec;
  private MarketParametersRecord marketParamRec;
  private DataParametersRecord dataParamRec;

  private GridBagLayout gridbag;
  private JPanel lpanel, cpanel, c1panel, c2panel, rpanel, buttonpanel, radiobuttonpanel;

  private JRadioButton chenPriceRadioButton, excessDemmandPriceRadioButton;
  private ButtonGroup  pricesGroup = new ButtonGroup();

  private Scroller niscroller;
  private Scroller tpscroller;
  private Scroller ftscroller;
  private Scroller ntscroller;
  private Scroller ttscroller;
  private Scroller gnscroller;
  private Scroller popscroller;
  private Scroller genscroller;
  private Scroller parscroller;
  private Scroller maxdscroller;
  private Scroller pcdscroller;
  private Scroller pmscroller;

  private TexFld initialDateLabel;
  private TexFld finalDateLabel;
  private TexFld fileNameLabel;
  private TexFld betha1Label;
  private TexFld betha2Label;
  private TexFld marketDepthLabel;
  private TexFld seedLabel;
  private TexFld delayLabel;
  private TexFld savingFreqLabel;

  private Button okbutton;
  private Button cancelbutton;
  private Button loadButton;
  private Button editTechTradersButton ;
  private Button editFundamentalTradersButton;
  private Button editNoisyTradersButton;
  private Market market;
  private StockMarketApplet applet;

  private OptTechnicalFrame opttechnicalframe;
  private OptGroupNumFrame optgroupnumframe;
  private OptFundamentalFrame optfundamentalframe;
  private OptNoisyFrame optnoisyframe;

  // Create an instance of OptFrame, using the provided title, the Market object
  // whose value are to be changed and a reference to the parent Applet
  public OptFrame(String title, Market market, StockMarketApplet app)
  {
    super(title);
    this.setBackground(Color.lightGray);
    this.market = market;
    this.applet = app;
    gridbag= new GridBagLayout();
    setLayout(gridbag);
    lpanel= new JPanel();
    lpanel.setLayout(gridbag);
    lpanel.setBackground(Color.lightGray);
    lpanel.setBorder(BorderFactory.createTitledBorder("Market Parameters"));
    cpanel= new JPanel();
    cpanel.setLayout(gridbag);
    cpanel.setBackground(Color.lightGray);
    c1panel= new JPanel();
    c1panel.setLayout(gridbag);
    c1panel.setBackground(Color.lightGray);
    c1panel.setBorder(BorderFactory.createTitledBorder("Data Parameters"));
    c2panel= new JPanel();
    c2panel.setLayout(gridbag);
    c2panel.setBackground(Color.lightGray);
    c2panel.setBorder(BorderFactory.createTitledBorder("Simulation Parameters"));
    rpanel= new JPanel();
    rpanel.setLayout(gridbag);
    rpanel.setBackground(Color.lightGray);
    rpanel.setBorder(BorderFactory.createTitledBorder("Traders"));
    buttonpanel= new JPanel();
    buttonpanel.setLayout(gridbag);
    buttonpanel.setBackground(Color.lightGray);
    radiobuttonpanel = new JPanel();
    radiobuttonpanel.setLayout(gridbag);
    radiobuttonpanel.setBackground(Color.lightGray);
    radiobuttonpanel.setBorder(BorderFactory.createTitledBorder("Price Equation"));

    GridBagConstraints c = new GridBagConstraints();
    Util util = new Util();

    this.readParameters();

    niscroller= new Scroller("Number Iterations", 1, 10010, marketParamRec.getMaxNumTradingPeriods());
    util.constraints(lpanel, niscroller, c, 0, 0, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    tpscroller = new Scroller("Porportion to trade", 1, 110, marketParamRec.getTradingProportion());
    util.constraints(lpanel, tpscroller, c, 0, 1, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    chenPriceRadioButton = new JRadioButton("P(t)=P(t-1)(1+betha(Bt-Ot))");
    chenPriceRadioButton.setBackground(Color.lightGray);
    chenPriceRadioButton.setMnemonic(KeyEvent.VK_C);
    chenPriceRadioButton.setActionCommand("P(t)=P(t-1)(1+betha(Bt-Ot))");
    chenPriceRadioButton.setSelected(this.marketParamRec.getPriceEquation()==CHEN);
    excessDemmandPriceRadioButton = new JRadioButton("P(t)=P(t-1)+((Excess Demand)/NumberTraders*c)");
    excessDemmandPriceRadioButton.setBackground(Color.lightGray);
    excessDemmandPriceRadioButton.setMnemonic(KeyEvent.VK_C);
    excessDemmandPriceRadioButton.setActionCommand("P(t)=P(t-1)+((Excess Demand)/NumberTraders*c)");
    excessDemmandPriceRadioButton.setSelected(this.marketParamRec.getPriceEquation()==EXCESS);
    pricesGroup.add(chenPriceRadioButton);
    pricesGroup.add(excessDemmandPriceRadioButton);

    util.constraints(radiobuttonpanel, chenPriceRadioButton, c, 0, 1, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);

    betha1Label = new TexFld ("Betha 1",""+(marketParamRec.getBetha1()),8);
    betha1Label.setEnabled(this.marketParamRec.getPriceEquation()==CHEN);
    util.constraints(radiobuttonpanel, betha1Label, c, 0, 2, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    betha2Label = new TexFld ("Betha 2",""+(marketParamRec.getBetha2()),8);
    betha2Label.setEnabled(this.marketParamRec.getPriceEquation()==CHEN);
    util.constraints(radiobuttonpanel, betha2Label, c, 0, 3, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    util.constraints(radiobuttonpanel, excessDemmandPriceRadioButton, c, 0, 4, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    marketDepthLabel = new TexFld ("Market Depth Constant c",""+(marketParamRec.getMarketDepth()),4);
    marketDepthLabel.setEnabled(this.marketParamRec.getPriceEquation()==EXCESS);
    util.constraints(radiobuttonpanel, marketDepthLabel, c, 0, 5, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    util.constraints(lpanel, radiobuttonpanel, c, 0, 2, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);

    initialDateLabel = new TexFld("Initial Date", dataParamRec.getInitialDate(),8);
    util.constraints(c1panel, initialDateLabel, c, 0,0,GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    finalDateLabel = new TexFld("Final Date", dataParamRec.getFinalDate(),8);
    util.constraints(c1panel, finalDateLabel, c, 0,1,GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    fileNameLabel = new TexFld("File Name", dataParamRec.getFileNameString(),33);
    util.constraints(c1panel, fileNameLabel, c, 0,2,GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);

    seedLabel = new TexFld("Simulation Seed", ""+(marketParamRec.getSeed()),8);
    util.constraints(c2panel, seedLabel, c, 0, 0, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    delayLabel = new TexFld("Screen delay",""+(marketParamRec.getDelay()),8);
    util.constraints(c2panel, delayLabel, c, 0, 1,  GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    savingFreqLabel = new TexFld("Saving frequency", ""+(marketParamRec.getSavingFrequency()),8);
    util.constraints(c2panel, savingFreqLabel, c, 0, 2,  GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);

    ttscroller= new Scroller("Number of groups of technical traders", 0, 510, this.tradersParamRec.getNumGroupsTechnicalTraders());
    util.constraints(rpanel, ttscroller, c, 0, 0, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    editTechTradersButton = new Button("Edit Technical Traders Group Parameterss");
    c.insets = new Insets(10,0,0,0);
    util.constraints(rpanel, editTechTradersButton, c, 0, 1, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    ftscroller= new Scroller("Number of fundamental traders", 0, 510, this.tradersParamRec.getNumFundamentalTraders());
    util.constraints(rpanel, ftscroller, c, 0, 2, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    editFundamentalTradersButton = new Button("Edit Fundamental Traders Parameters");
    c.insets = new Insets(10,0,0,0);
    util.constraints(rpanel, editFundamentalTradersButton, c, 0, 3, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    ntscroller= new Scroller("Number of noisy traders", 0, 510, this.tradersParamRec.getNumNoisyTraders());
    util.constraints(rpanel, ntscroller, c, 0, 4, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);
    editNoisyTradersButton = new Button("Edit Noisy Traders Parameters");
    c.insets = new Insets(10,0,0,0);
    util.constraints(rpanel, editNoisyTradersButton, c, 0, 5, GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1, 1, 0.0, 0.0);

    util.constrain(buttonpanel, okbutton= new Button("Ok"), 1, 0, 1, 1, 5, 10, 0, 0);
    util.constrain(buttonpanel, cancelbutton= new Button("Cancel"), 2, 0, 1, 1, 5, 10, 0, 0);

    util.constrain(this, new Label("Parameter options"), 0, 0, 6, 1,
	      GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1.0, 0.0, 0, 400, 0, 400);
    util.constrain(this, lpanel, 0, 1, 1, 1,
	      GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1.0, 0.0, 0, 0, 0, 0);
    util.constrain(cpanel, c1panel, 1, 1, 1, 1,
	      GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1.0, 0.0, 0, 0, 0, 0);
    util.constrain(cpanel, c2panel, 1, 2, 1, 1,
	      GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1.0, 0.0, 0, 0, 0, 0);
    util.constrain(this,cpanel,  1, 1, 1, 1,
              GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1.0, 0.0, 0, 0, 0, 0);
    util.constrain(this, rpanel, 2, 1, 1, 1,
	      GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1.0, 0.0, 0, 0, 0, 0);
    util.constrain(this, buttonpanel, 0, 2, 3, 1,
	      GridBagConstraints.HORIZONTAL, GridBagConstraints.CENTER, 1.0, 1.0, 0, 0, 0, 0);
    pack();
  }

  // Handle input events
  public boolean action(Event e, Object arg)
  {
    if(e.target == okbutton)
    {
      saveParameters();
      this.dispose();
      applet.resetAll();
      return true;
    } else if(e.target == cancelbutton)
    {
      this.dispose();
      return true;
    } else if(e.target == editTechTradersButton)
    {
      if(optgroupnumframe == null){
        optgroupnumframe = new OptGroupNumFrame("Number of Group of Technical Traders to Edit", market, this, ttscroller.getValue());
        optgroupnumframe.setVisible(true);
      } else{
        optgroupnumframe.dispose();
        optgroupnumframe = new OptGroupNumFrame("Number of Group of Technical Traders to Edit", market, this, ttscroller.getValue());
        optgroupnumframe.setVisible(true);
      }
    } else if(e.target == editFundamentalTradersButton)
    {
      if(optfundamentalframe == null){
        optfundamentalframe = new OptFundamentalFrame("Parameters for Fundamental Traders", market, this);
        optfundamentalframe.setVisible(true);
      } else{
        optfundamentalframe.dispose();
        optfundamentalframe = new OptFundamentalFrame("Parameters for Fundamental Traders", market, this);
        optfundamentalframe.setVisible(true);
      }
    } else if(e.target == editNoisyTradersButton)
    {
      if(optnoisyframe == null){
        optnoisyframe = new OptNoisyFrame("Parameters for Noisy Traders", market, this);
        optnoisyframe.setVisible(true);
      } else{
        optnoisyframe.dispose();
        optnoisyframe = new OptNoisyFrame("Parameters for Noisy Traders", market, this);
        optnoisyframe.setVisible(true);
      }
    } else{
      return false;
    }
    return true;
  }

  private void readMarketParameters(){
    marketParamRec = new MarketParametersRecord();
    InputMarketParameters inputMarketParam =  new InputMarketParameters();
    marketParamRec = inputMarketParam.readParameters();
  }

  private void readTradersParameters(){
    tradersParamRec = new TradersParametersRecord ();
    InputTradersParameters inputTradersParam = new InputTradersParameters();
    tradersParamRec = inputTradersParam.readParameters();
  }

  private void readDataParameters(){
    dataParamRec = new DataParametersRecord();
    InputDataParameters inputDataParam = new InputDataParameters();
    dataParamRec = inputDataParam.readParameters();
  }

  private void readParameters(){
    readMarketParameters();
    readTradersParameters();
    readDataParameters();
  }

  private void saveTradersParameters(){
    tradersParamRec.setNumGroupsTechnicalTraders(ttscroller.getValue());
    tradersParamRec.setNumFundamentalTraders(ftscroller.getValue());
    tradersParamRec.setNumNoisyTraders(ntscroller.getValue());
    OutputTradersParameters outputTradersParam = new OutputTradersParameters();
    outputTradersParam.saveParameters(this.tradersParamRec);
  }

  private void saveMarketParameters(){
    marketParamRec.setMaxNumTradingPeriods(niscroller.getValue());
    marketParamRec.setTradingProportion(tpscroller.getValue());
    if(chenPriceRadioButton.isSelected()){
      marketParamRec.setPriceEquation(CHEN);
      marketParamRec.setBetha1(new Double(betha1Label.getValue()).doubleValue() );
      marketParamRec.setBetha2(new Double(betha2Label.getValue()).doubleValue());
    }
    else{
      marketParamRec.setPriceEquation(EXCESS);
      marketParamRec.setMarketDepth(new Double(marketDepthLabel.getValue()).doubleValue());
    }
    marketParamRec.setSeed(new Long(seedLabel.getValue()).longValue());
    marketParamRec.setDelay(new Integer(delayLabel.getValue()).intValue());
    marketParamRec.setSavingFrequency(new Integer(savingFreqLabel.getValue()).intValue());
    OutputMarketParameters outputMarketParam = new OutputMarketParameters();
    outputMarketParam.saveParameters(marketParamRec);
  }

  private void saveDataParameters(){
    dataParamRec.setInitialDate(initialDateLabel.getValue());
    dataParamRec.setFinalDate(finalDateLabel.getValue());
    dataParamRec.setFileNameString(fileNameLabel.getValue());
    OutputDataParameters outputDataParam = new OutputDataParameters();
    outputDataParam.saveParameters(dataParamRec);
  }

  private void saveParameters(){
    saveTradersParameters();
    saveMarketParameters();
    saveDataParameters();
  }

}