package stockMarketApplet.market;

import java.io.*;
import java.awt.*;
import java.lang.Math;
import java.util.List;
import java.util.Random;
import java.util.Calendar;
import java.util.ArrayList;
import stockMarketApplet.util.*;
import java.util.GregorianCalendar;
import stockMarketApplet.market.io.*;
import stockMarketApplet.market.agents.*;
import stockMarketApplet.StockMarketApplet;
import stockMarketApplet.market.agents.technicalTrader.*;
import stockMarketApplet.market.agents.technicalTrader.basics.*;

public class Market extends Object implements Runnable {
  public final static int BEST  = 0;
  public final static int NEXT  = 1;
  public final static int PREV  = 2;
  public final static int WORST = 3;
  private static final byte BUY  = 1;
  private static final byte SELL = -1;
  private static final byte HOLD = 0;
  private static final int CHEN   = 0;
  private static final int EXCESS = 1;
  private static final int TECHNICAL    = 0;
  private static final int FUNDAMENTAL  = 1;
  private static final int NOISY        = 2;
  private static final byte DONT_GENERATE_LIMIT_ORDER_SELL = 0;
  private static final byte GENERATE_LIMIT_ORDER_SELL = 1;
  private static final byte DONT_GENERATE_LIMIT_ORDER_BUY = 0;
  private static final byte GENERATE_LIMIT_ORDER_BUY = 1;
  private static final int MV_12_Window       = 12;
  private static final int MV_50_Window       = 50;
  private static final int TRB_5_Window       = 5;
  private static final int TRB_50_Window      = 50;
  private static final int Filter_5_Window    = 5;
  private static final int Filter_63_Window   = 63;
  private static final int VOL_12_Window      = 12;
  private static final int VOL_50_Window      = 50;
  private static final int MOM_10_Window      = 10;
  private static final int MOM_60_Window      = 60;
  private static final int MMA_10_Window      = 10;
  private static final int MMA_60_Window      = 60;
  public static final int DRAWEX  = 0;
  public static final int DRAWIMP = 1;
  private static final int IDLE = 0;
  private static final int RUN  = 1;
  private static final int STOP = 2;
  private static final int FIXED    = 0;
  private static final int REDQUEEN = 1;
  private static String directoryName;
  private static Random rand;
  private static Util util;
  private static FileOutputStream fileNameTT, fileNameFT, fileNameNT;
  private static TechnicalTrader[] technicalTraders;
  private static TechnicalTrader[] replicatedTraders;
  private static TechnicalTrader[] sortedTechTraders;
  private static FundamentalTrader[] fundamentalTraders;
  private static NoisyTrader[] noisyTraders;
  private static InputData inputData;
  private static int selected;
  private static double fundamentalValue;
  private static TradingRecord[] tradingRecords;
  private static GPParametersRecord[] gPParamRec;
  private static IndicatorsRecord[] indicatorsRec;
  private static FunctionsRecord[] relFunctionsRec;
  private static DataParametersRecord dataParamRec;
  private static FunctionsRecord[] logicFunctionsRec;
  private static MarketParametersRecord marketParamRec;
  private static DataRecord[] trainingData, tradingData;
  private static GroupParametersRecord[] groupsParamRec;
  private static TradersParametersRecord tradersParamRec;
  private static FundamentalParametersRecord fundamentalParamRec;
  private static NoisyTraderParametersRecord noisyTradersParamRec;

  private static int numberTechnicalTraders, numberNoisyTraders, numberFundamentalTraders;
  private static int totalNumShares, numberTrainingData, tradingPeriod;

  private static int tmp=0,counter=0,maxMemTrainingData=0;

  //  flag to decide if it is necesary to execute the initial training
  private static boolean loaded = false;
  // flag to decide if current individual shall be drawn
  protected boolean drawit;
  // Graphics context of the assigned Applet
  protected Graphics graph;
  // bound of the Canvas the individual shall be drawn in
  protected Rectangle bounds;
  // Link to the Applet this Market belongs to. Needed for callbacks
  protected static StockMarketApplet parent_app;
  // Thread that evaluates the Individuals and creates new generations
  protected Thread market = null;
  // State variable for the computation thread
  private static Integer runStatus;
  // Market's constructor
  public Market(){
    runStatus = new Integer(IDLE);
    selected = BEST;
  }
  //  Method that reads the market parameters
  private static void readMarketParameters(){
    marketParamRec = new MarketParametersRecord();
    InputMarketParameters inputMarketParam = new InputMarketParameters() ;
    marketParamRec = inputMarketParam.readParameters();
  }
  //  Method that reads the traders parameters
  private static void readTradersParameters(){
    tradersParamRec = new TradersParametersRecord();
    InputTradersParameters inputTradersParam = new InputTradersParameters();
    tradersParamRec = inputTradersParam.readParameters();
  }
  //  Mathod that reads the groups parameters
  private static void readGroupsParameters(){
    groupsParamRec = new GroupParametersRecord[tradersParamRec.getNumGroupsTechnicalTraders()];
    InputGroupParameters inputGroupsParam = new InputGroupParameters() ;
    groupsParamRec = inputGroupsParam.readGroupsGroupParameters(tradersParamRec.getNumGroupsTechnicalTraders());
  }
  //  Method that reads the data parameters
  private static void readDataParameters(){
    dataParamRec = new DataParametersRecord() ;
    InputDataParameters inputDataParam = new InputDataParameters() ;
    dataParamRec=inputDataParam.readParameters();
  }
  //  Method that reads the GP parameters
  private static void readGPParameters(){
    gPParamRec = new GPParametersRecord[tradersParamRec.getNumGroupsTechnicalTraders()];
    InputGPParameters inputGPParam = new InputGPParameters();
    gPParamRec = inputGPParam.readGroupsGPParameters(tradersParamRec.getNumGroupsTechnicalTraders());
  }
  //  Method that read the logical functions
  private static void readLogicFunctions(){
    logicFunctionsRec = new FunctionsRecord[tradersParamRec.getNumGroupsTechnicalTraders()];
    InputFunctionsLogic inputFunctLogic = new InputFunctionsLogic();
    logicFunctionsRec = inputFunctLogic.readGroupsFunctions(tradersParamRec.getNumGroupsTechnicalTraders());
  }
  //  Method that reads the relational functions
  private static void readRelFunctions(){
    relFunctionsRec = new FunctionsRecord[tradersParamRec.getNumGroupsTechnicalTraders()];
    InputFunctionsRel inputFunctRel = new InputFunctionsRel();
    relFunctionsRec = inputFunctRel.readGroupsFunctions(tradersParamRec.getNumGroupsTechnicalTraders());
  }
  //  Method that reads the indicators
  private static void readIndicators(){
    indicatorsRec = new IndicatorsRecord[tradersParamRec.getNumGroupsTechnicalTraders()];
    InputIndicators inputIndicators = new InputIndicators();
    indicatorsRec = inputIndicators.readGroupsIndicators(tradersParamRec.getNumGroupsTechnicalTraders());
  }
  private static void readNoisyTradersParameters(){
    noisyTradersParamRec = new NoisyTraderParametersRecord();
    InputNoisyTradersParameters inputNoisyTradersParam = new InputNoisyTradersParameters();
    noisyTradersParamRec = inputNoisyTradersParam.readParameters();
  }
  private static void readFundamentalTradersParameters(){
    fundamentalParamRec = new FundamentalParametersRecord();
    InputFundamentalParameters inputFundamentalParam = new InputFundamentalParameters();
    fundamentalParamRec = inputFundamentalParam.readParameters();
  }
  //  Method that calculates the total number of technical traders
  private static int computeNumTechnicalTraders(){
    int temp=0;
    for (int i = 0; i < tradersParamRec.getNumGroupsTechnicalTraders(); i++ ){
      temp+=groupsParamRec[i].getNumTraders();
    }
    return temp;
  }
  //  Method that calculates the total number of shares held by the traders
  private static int computeTotalNumShares(){
    int temp = 0;
    for (int i=0;i<tradersParamRec.getNumGroupsTechnicalTraders();i++){
      temp+=(groupsParamRec[i].getNumberShares()*groupsParamRec[i].getNumTraders());
    }
    temp += ( numberNoisyTraders*((int) noisyTradersParamRec.getNumberShares()) );
    temp += ( numberFundamentalTraders*((int) fundamentalParamRec.getNumberShares()) );
    return temp;
  }
  //  After reading the logical functions from the files, creates the function nodes
  private static List createLogicFunct(String[] functionsNames){
    List tempLogicFunctions = new ArrayList();
    for (int i = 0; i  < functionsNames.length; i++){
      tempLogicFunctions.add(functionsNames[i]);
    }
    return tempLogicFunctions;
  }
  //  After reading the relational functions from the files, creates the function nodes
  private static List createRelFunct(String[] functionsNames){
    List tempRelFunctions = new ArrayList();
    for (int i = 0; i  < functionsNames.length; i++){
      tempRelFunctions.add(functionsNames[i]);
    }
    return tempRelFunctions;
  }
  //  After reading the indicators from the files, creates the indicator nodes
  private static List createIndicators(String[] indicatorsNames){
    List tempIndicators = new  ArrayList();
    for (int i = 0; i  < indicatorsNames.length; i++){
      tempIndicators.add(indicatorsNames[i]);
    }
    return tempIndicators;
  }
  //  Creates the tecnical traders
  private static void createTechnicalTraders(){
    int index = 0;
    double tempT, tempTDiff, tempTau, tempTauDiff;
    tempTDiff = fundamentalParamRec.getTmax()-fundamentalParamRec.getTmin();
    tempTauDiff = fundamentalParamRec.getTauMax()-fundamentalParamRec.getTauMin();
    technicalTraders = new TechnicalTrader[numberTechnicalTraders];
    inputData = new InputData(dataParamRec.getInitialDate(), dataParamRec.getFinalDate(),dataParamRec.getFileNameString());
    for (int i = 0; i < tradersParamRec.getNumGroupsTechnicalTraders(); i++ ){
          for (int j = 0; j < groupsParamRec[i].getNumTraders(); j++ ){
            int[] tempRecommend = new int[numberTrainingData];
            maxMemTrainingData = Math.max(groupsParamRec[i].getMemoryLenght(),numberTrainingData);
            tempRecommend = inputData.fillDecisions(groupsParamRec[i].getRate(), groupsParamRec[i].getDays(), trainingData);
            technicalTraders[index] = new TechnicalTrader( groupsParamRec[i], gPParamRec[i], maxMemTrainingData, tempRecommend,
                                                  createIndicators(indicatorsRec[i].getIndicatorsNames()),
                                                  createLogicFunct(logicFunctionsRec[i].getFunctsNames()),
                                                  createRelFunct(relFunctionsRec[i].getFunctsNames()));

            tempT = fundamentalParamRec.getTmin() + (tempTDiff*rand.nextDouble());
            tempTau = fundamentalParamRec.getTauMin() + (tempTauDiff*rand.nextDouble());
            technicalTraders[index].setT(tempT);
            technicalTraders[index].setTau(tempTau);
            index++;
          }
    }
  }
  private static void createNoisyTraders(){
    noisyTraders = new NoisyTrader[numberNoisyTraders];
    for (int i=0; i<numberNoisyTraders; i++){
      noisyTraders[i] = new NoisyTrader(noisyTradersParamRec);
    }
  }
  private static void createFundamentalTraders(){
    fundamentalTraders = new FundamentalTrader[numberFundamentalTraders];
    double tempT, tempTDiff, tempTau, tempTauDiff;
    tempTDiff = fundamentalParamRec.getTmax()-fundamentalParamRec.getTmin();
    tempTauDiff = fundamentalParamRec.getTauMax()-fundamentalParamRec.getTauMin();
    for (int i=0; i<numberFundamentalTraders; i++){
      tempT = fundamentalParamRec.getTmin() + (tempTDiff*rand.nextDouble());
      tempTau = fundamentalParamRec.getTauMin() + (tempTauDiff*rand.nextDouble());
      fundamentalTraders[i] = new FundamentalTrader(fundamentalParamRec);
      fundamentalTraders[i].setT(tempT);
      fundamentalTraders[i].setTau(tempTau);
    }
  }
  //  Initiates the market
  public static void init(){
    readMarketParameters();
    readTradersParameters();
    readGroupsParameters();
    readDataParameters();
    readGroupsParameters();
    readGPParameters();
    readLogicFunctions();
    readRelFunctions();
    readIndicators();
    readNoisyTradersParameters();
    readFundamentalTradersParameters();
    rand  = new Random();
    rand.setSeed(marketParamRec.getSeed());
    util = new Util();
    InputData inputData = new InputData(dataParamRec.getInitialDate(), dataParamRec.getFinalDate(),dataParamRec.getFileNameString());
    trainingData = inputData.loadData(dataParamRec.getFileNameString());
    numberTrainingData = trainingData.length;
    numberTechnicalTraders = computeNumTechnicalTraders();
    sortedTechTraders = new TechnicalTrader[numberTechnicalTraders];
    numberNoisyTraders = tradersParamRec.getNumNoisyTraders();
    numberFundamentalTraders = tradersParamRec.getNumFundamentalTraders();
    totalNumShares = computeTotalNumShares();
    createTechnicalTraders();
    createFundamentalTraders();
    createNoisyTraders();
    Calendar cal = new GregorianCalendar();
    int year = cal.get(Calendar.YEAR);             // 2002
    int month = cal.get(Calendar.MONTH)+1;         // 0=Jan, 1=Feb, ...
    int day = cal.get(Calendar.DAY_OF_MONTH);      // 1...
    int hour24 = cal.get(Calendar.HOUR_OF_DAY);     // 0..23
    int min = cal.get(Calendar.MINUTE);             // 0..59
    directoryName = ""+year+"-"+month+"-"+day+"."+hour24+min;
    boolean success = (new File(directoryName)).mkdir();
    if (!success) {
      // Directory creation failed
      parent_app.printStatus("Failure to create directory");
    }
    util.copyParametersFiles(directoryName);
    tradingData     = new DataRecord[Filter_63_Window + marketParamRec.getMaxNumTradingPeriods()];
    tradingRecords  = new TradingRecord[Filter_63_Window + marketParamRec.getMaxNumTradingPeriods()];
    //  Copy the first necesary elements to allow the computation of the indicators
    for (int i = 0; i < Filter_63_Window; i++ ){
      tradingData[i] = trainingData[trainingData.length - Filter_63_Window + i];
    }
  }
  // Method that executes the initial training
  public static void initialTraining(){
    init();
    Calendar cal = new GregorianCalendar();
    System.out.println(cal.get(Calendar.HOUR_OF_DAY)+":"+cal.get(Calendar.MINUTE)+"."+cal.get(Calendar.SECOND)+"."+cal.get(Calendar.MILLISECOND));
    for (int i = 0; i < numberTechnicalTraders; i++ ){
      technicalTraders[i].setWealth(technicalTraders[i].getCash() + (technicalTraders[i].getStocks() * tradingData[Filter_63_Window-1].getObservation()) );
      parent_app.printStatus("Initial training Trader: "+i);
      technicalTraders[i].initialTraining(trainingData);
    }
    cal = new GregorianCalendar();
    System.out.println(cal.get(Calendar.HOUR_OF_DAY)+":"+cal.get(Calendar.MINUTE)+"."+cal.get(Calendar.SECOND)+"."+cal.get(Calendar.MILLISECOND));
  }
  //  Method that load previous training
  public static void loadInitialTraining(){
    init();
    int k = 0;
    for (int i = 0; i < groupsParamRec.length; i++ ){
      for (int j=0;j<groupsParamRec[i].getNumTraders();j++){
        parent_app.printStatus("Loading previous training Trader: "+j+" Group: "+i);
        technicalTraders[k].loadPopulation(i,j);
        k++;
      }
    }
      loaded = true;
  }
  //  Method that saves the currents Technical Traders' populations
  public static void saveInitialTraining(){
    int k = 0;
    for (int i = 0; i < groupsParamRec.length; i++ ){
      for (int j=0;j<groupsParamRec[i].getNumTraders();j++)   {
        parent_app.printStatus("Saving training Trader: "+j+" Group: "+i);
        technicalTraders[k].savePopulation(i,j);
        k++;
      }
    }
  }
  //  Method that calculates the indicators
  private static void computeIndicators(int index){
    double temp;
    DataRecord[] tempData;
    tradingData[index].setDate( new Integer(index - Filter_63_Window).toString() );
    tempData = new DataRecord[MV_12_Window];
    tempData = util.subarray(tradingData, (index - MV_12_Window) + 1, index);
    temp = (tradingData[index].getObservation() - util.computeAverage(tempData)) / (util.computeAverage(tempData));
    tradingData[index].setMV_12(temp);
    tempData = null;
    tempData = new DataRecord[MV_50_Window];
    tempData = util.subarray(tradingData, (index - MV_50_Window) + 1, index);
    temp = (tradingData[index].getObservation() - util.computeAverage(tempData)) / (util.computeAverage(tempData));
    tradingData[index].setMV_50(temp);
    tempData = null;
    tempData = new DataRecord[TRB_5_Window];
    tempData = util.subarray(tradingData, (index - TRB_5_Window), index - 1);
    temp = (tradingData[index].getObservation() - util.computeMax(tempData)) / (util.computeMax(tempData));
    tradingData[index].setTRB_5(temp);
    tempData = null;
    tempData = new DataRecord[TRB_50_Window];
    tempData = util.subarray(tradingData, (index - TRB_50_Window), index - 1);
    temp = (tradingData[index].getObservation() - util.computeMax(tempData)) / (util.computeMax(tempData));
    tradingData[index].setTRB_50(temp);
    tempData = null;
    tempData = new DataRecord[Filter_5_Window];
    tempData = util.subarray(tradingData, (index - Filter_5_Window), index - 1);
    temp = (tradingData[index].getObservation() - util.computeMin(tempData)) / (util.computeMin(tempData));
    tradingData[index].setFilter_5(temp);
    tempData = null;
    tempData = new DataRecord[Filter_63_Window];
    tempData = util.subarray(tradingData, (index - Filter_63_Window), index - 1);
    temp = (tradingData[index].getObservation() - util.computeMin(tempData)) / (util.computeMin(tempData));
    tradingData[index].setFilter_63(temp);
    tempData = null;
    tempData = new DataRecord[VOL_12_Window];
    tempData = util.subarray(tradingData, (index - VOL_12_Window) + 1, index);
    temp = (util.computeStdDev(tempData)) / (util.computeAverage(tempData));
    tradingData[index].setVOL_12(temp);
    tempData = null;
    tempData = new DataRecord[VOL_50_Window];
    tempData = util.subarray(tradingData, (index - VOL_50_Window) + 1, index);
    temp = (util.computeStdDev(tempData)) / (util.computeAverage(tempData));
    tradingData[index].setVOL_50(temp);
    tempData = null;
    temp = (tradingData[index].getObservation()-tradingData[index-MOM_10_Window].getObservation())/10.0;
    tradingData[index].setMOM_10(temp);
    tempData = null;
    temp = (tradingData[index].getObservation()-tradingData[index-MOM_60_Window].getObservation())/10.0;
    tradingData[index].setMOM_60(temp);
    tempData = null;
    tempData = new DataRecord[MMA_10_Window];
    tempData = util.subarray(tradingData, (index - MMA_10_Window), index - 1);
    temp = (tradingData[index].getMOM_10() - util.computeAverage(tempData,"MOM_10")) / (util.computeAverage(tempData,"MOM_10"));
    tradingData[index].setMMA_10(temp);
    tempData = null;
    tempData = new DataRecord[MMA_60_Window];
    tempData = util.subarray(tradingData, (index - MMA_60_Window), index - 1);
    temp = (tradingData[index].getMOM_60() - util.computeAverage(tempData,"MOM_60")) / (util.computeAverage(tempData,"MOM_60"));
    tradingData[index].setMMA_60(temp);
  }
  //////////////////////////////////////////////////////////
  //  This is the method to assing shares to each trader  //
  //////////////////////////////////////////////////////////
  private static void updateCashAndStocks(int index){
    double totalBids = 0.0, totalOffers = 0.0, m_totalBids = 0.0;
    double previousPrice = tradingData[index-1].getObservation(), nextPrice;
    double tradingProportion = (marketParamRec.getTradingProportion()/100.0);
    short tempShort = 0;
    byte tempDes = HOLD;
    int numberOfferers = 0, numberBiders = 0, numberHolders = 0, numPendingSellsOrBuys = 0,tmpNumSellers =0,tmpNumBuyers=0;
    double avgTotalBids = 0.0, avgTotalOffers = 0.0, avgDiffOffers = 0.0;
    double stocks = 0.0, cash = 0.0, rate = 0.0;
    double tmpDiff = 0.0, tmpA = fundamentalParamRec.getA();
    double fracFilledBuyOrders = 0.0, fracFilledSellOrders = 0.0;
    int time = 0, opType = HOLD, date = 0, days = 0;
    double  tmpRnd=0.0;
    ////////////////////////////////////////////////////////////////
    //  Computation of the total number of shares to sell or buy  //
    ////////////////////////////////////////////////////////////////
    //  Bouchaud's trading proportion for NOISY TRADERS
    for (int i = 0; i < noisyTraders.length; i++ ){
      stocks = noisyTraders[i].getNumberShares();
      cash = noisyTraders[i].getCash();
      tmpRnd = rand.nextDouble();
      tempDes = noisyTraders[i].takeDesicion(tmpRnd);
      // The current noisy trader WANTS to SELL and has shares
      if ((tempDes==SELL) && (stocks>0)){
        totalOffers += (stocks*tradingProportion);
        numberOfferers++;
      }
      //  The current noisy trader WANTS to BUY and has cash
      else if ((tempDes==BUY) && (cash>0)){
        totalBids += ((cash*tradingProportion/previousPrice));
        numberBiders++;
      }
      else{
        numberHolders++;
      }
      tmp+=tempDes;
      counter++;
    }
    //  Farmer's FUNDAMENTAL TRADERS
    fundamentalValue = updateFundamentalValue(index);
    for (int i = 0; i < fundamentalTraders.length; i++ ){
      stocks = fundamentalTraders[i].getNumberShares();
      cash = fundamentalTraders[i].getCash();
      tmpDiff = fundamentalTraders[i].getDifference();
      // The current fundamental trader is in a SELLING cycle and has shares
      if (fundamentalTraders[i].isSelling() && (stocks>0)){
        totalOffers += Math.min(stocks,(tmpDiff*tmpA)/previousPrice);
        numberOfferers++;
      }
      //  The current fundamental trader is in a BUYING cycle and has cash
      else if(fundamentalTraders[i].isBuying() && (cash>0)){
        totalBids += Math.min((cash/previousPrice),(tmpDiff*tmpA)/previousPrice);
        numberBiders++;
      }
      else {
        tempDes = fundamentalTraders[i].takeDesicion(previousPrice,fundamentalValue);
        // The current fundamental trader WANTS to SELL and has shares
        if ((tempDes==SELL) && (stocks>0)){
          totalOffers += Math.min(stocks,(tmpDiff*tmpA)/previousPrice);
          numberOfferers++;
        }
        //  The current fundamental trader WANTS to BUY and has cash
        else if ((tempDes==BUY) && (cash>0)){
          totalBids += Math.min((cash/previousPrice),(tmpDiff*tmpA)/previousPrice);
          numberBiders++;
        }
      }
    }
    //  Bouchaud's trading proportion for TECHNICAL TRADERS and Farmer's FUNDAMENTAL BEHAVIOUR
    for (int i = 0; i < technicalTraders.length; i++ ){
      stocks = technicalTraders[i].getStocks();
      cash = technicalTraders[i].getCash();
      rate = technicalTraders[i].getRate();
      numPendingSellsOrBuys = technicalTraders[i].getLimitOrderBookSize();
      tmpDiff = technicalTraders[i].getDifference();
      //  Ask if the technical traders has no limit orders to be reviewed
      if (numPendingSellsOrBuys == 0){
        if (technicalTraders[i].getFundamentalBehaviour()){
          if (technicalTraders[i].isSelling()){
            totalOffers += Math.min(stocks,(tmpDiff*tmpA)/previousPrice);
            numberOfferers++;
          }
          else if(technicalTraders[i].isBuying()){
            totalBids += Math.min((cash/previousPrice),(tmpDiff*tmpA)/previousPrice);
            numberBiders++;
          }
          else{
            // Takes a decision based on the fundamental behaviour
            tempDes = technicalTraders[i].takeDesicion(previousPrice,fundamentalValue);
            // The current trader WANTS to SELL and has shares
            if ((tempDes==SELL) && (stocks>0)){
              totalOffers += Math.min(stocks,(tmpDiff*tmpA)/previousPrice);
              numberOfferers++;
            }
            //  The current trader WANTS to BUY and has cash
            else if ((tempDes==BUY) && (cash>0)){
              totalBids += Math.min((cash/previousPrice),(tmpDiff*tmpA)/previousPrice);
              numberBiders++;
            }
            else{
              // Make Prediction with the GEDEFIT
              tempDes = technicalTraders[i].getBestProgram().evaluation(tempShort,tradingData[index-1]);
              // The current technical trader WANTS to SELL and has shares
              if ((tempDes==SELL) && (stocks>0)){
                totalOffers += (stocks*tradingProportion);
                numberOfferers++;
              }
              //  The current technical trader WANTS to BUY and has cash
              else if ((tempDes==BUY) && (cash>0)){
                totalBids += (((cash*tradingProportion)/previousPrice));
                numberBiders++;
              }
            }// Else technical trader did not trigger her fundamental behaviour
          } // Else technical trader is not in a cycle of fundamental behaviour
        } // If technical trader has fundamental behaviour
        else{
          // Make Prediction with the GEDEFIT
          tempDes = technicalTraders[i].getBestProgram().evaluation(tempShort, tradingData[index-1]);
          // The current technical trader WANTS to SELL and has shares
          if ((tempDes==SELL) && (stocks>0)){
            totalOffers += (stocks*tradingProportion);
            numberOfferers++;
          }
          //  The current technical trader WANTS to BUY and has cash
          else if ((tempDes==BUY) && (cash>0)){
            totalBids += (((cash*tradingProportion)/previousPrice));
            numberBiders++;
          }
        } // Else technical trader has no fundamental behaviour
      } // If technical trader has no limit orders to be executed
      //  The current technical trader HAS to SELL and has shares
      else if (validLimitOrder(i,index-1,SELL) && (stocks>0)){
        LimitOrderBook tempLimitOrder = new LimitOrderBook();
        tempLimitOrder = technicalTraders[i].getLimitOrder(0);
        double limitOrderSells = tempLimitOrder.getNumShares();
        totalOffers += Math.min(limitOrderSells,stocks);
        numberOfferers++;
      } // If the tech trader has a limit order to sell
      //  The current technical trader HAS to BUY and has cash
      else if (validLimitOrder(i,index-1,BUY) && (cash>0)){
        LimitOrderBook tempLimitOrder = new LimitOrderBook();
        tempLimitOrder = technicalTraders[i].getLimitOrder(0);
        double limitOrderBuys = tempLimitOrder.getNumShares();
        if (cash > (limitOrderBuys*previousPrice)){
          totalBids += (limitOrderBuys);
        }
        else {
          totalBids += (cash/previousPrice);
        }
        numberBiders++;
      } // If the tech trader has a limit order to buy
    } // For cycle for technical traders
    tradingRecords[index] = new TradingRecord();
    tradingRecords[index].setEpoch(index - Filter_63_Window);
    tradingRecords[index].setBids(totalBids);
    tradingRecords[index].setOffers(totalOffers);
    tradingRecords[index].setNumSellers(numberOfferers);
    tradingRecords[index].setNumBuyers(numberBiders);
    tradingRecords[index].setNumHolders(numberHolders);
    nextPrice = updatePrice(index);
    tradingRecords[index].setFundamentalValue(fundamentalValue);
    m_totalBids = totalBids * (previousPrice/nextPrice);
    fracFilledBuyOrders = Math.min(1.0,(totalOffers/m_totalBids));
    fracFilledSellOrders = Math.min(1.0,(m_totalBids/totalOffers));
    //////////////////////////////////////////////////////
    //  Bouchaud's Rationing Scheme for Market Clearing //
    //////////////////////////////////////////////////////
    // Rationing for NOISY TRADERS
    for (int i = 0; i < noisyTraders.length; i++ ){
      stocks = noisyTraders[i].getNumberShares();
      cash = noisyTraders[i].getCash();
      tempDes = noisyTraders[i].getDecision();
      // The current trader WANTS to SELL
      if(tempDes == SELL){
        // The current trader has shares and there are traders willing to buy
        if ((stocks > 0)&&(totalBids > 0)){
          noisyTraders[i].setNumberShares( stocks - ((stocks*tradingProportion)*fracFilledSellOrders) );
          noisyTraders[i].setCash(cash + (nextPrice*((stocks*tradingProportion)*fracFilledSellOrders)));
        }
      }
      // The current trader WANTS to BUY
      else if (tempDes == BUY){
        // The current trader has cash and there are traderes willing to sell
        if ((cash > 0)&&(totalOffers > 0)){
          noisyTraders[i].setNumberShares( stocks + ((tradingProportion*(cash/nextPrice))*fracFilledBuyOrders) );
          noisyTraders[i].setCash(cash - (((tradingProportion*(cash/nextPrice))*fracFilledBuyOrders)*nextPrice));
        }
      }
    }
    //  Rationing for the FUNDAMENTAL TRADERS
    for (int i = 0; i < fundamentalTraders.length; i++ ){
      stocks = fundamentalTraders[i].getNumberShares();
      cash = fundamentalTraders[i].getCash();
      tmpDiff = fundamentalTraders[i].getDifference();
      // The current fundamental trader WANTS to SELL and has shares
      if (fundamentalTraders[i].isSelling() && (stocks>0)){
        fundamentalTraders[i].setNumberShares( stocks - ((Math.min(stocks, (tmpDiff*tmpA)/nextPrice))*fracFilledSellOrders) );
        fundamentalTraders[i].setCash(cash + (nextPrice*((Math.min(stocks,(tmpDiff*tmpA)/nextPrice))*fracFilledSellOrders)) );
        if (fundamentalTraders[i].exitPosition(nextPrice,fundamentalValue)){
          fundamentalTraders[i].setDecision(HOLD);
        }
      }
      //  The current fundamental trader WANTS to BUY and has cash
      else if(fundamentalTraders[i].isBuying() && (cash>0)){
        fundamentalTraders[i].setNumberShares( stocks + (Math.min((cash/nextPrice),(tmpDiff*tmpA)/nextPrice)*fracFilledBuyOrders) );
        fundamentalTraders[i].setCash(cash - (nextPrice*(Math.min((cash/nextPrice),(tmpDiff*tmpA)/nextPrice)*fracFilledBuyOrders)));
        if (fundamentalTraders[i].exitPosition(nextPrice,fundamentalValue)){
          fundamentalTraders[i].setDecision(HOLD);
        }
      }
      else {
        tempDes = fundamentalTraders[i].getDecision();
        // The current fundamental trader WANTS to SELL and has shares
        if ((tempDes==SELL) && (stocks>0)){
          fundamentalTraders[i].setNumberShares( stocks - ((Math.min(stocks,(tmpDiff*tmpA)/nextPrice))*fracFilledSellOrders) );
          fundamentalTraders[i].setCash(cash + (nextPrice*((Math.min(stocks,(tmpDiff*tmpA)/nextPrice))*fracFilledSellOrders)));
          fundamentalTraders[i].setSelling(true);
        }
        //  The current fundamental trader WANTS to BUY and has cash
        else if ((tempDes==BUY) && (cash>0)){
          fundamentalTraders[i].setNumberShares( stocks + (Math.min((cash/nextPrice),(tmpDiff*tmpA)/nextPrice)*fracFilledBuyOrders) );
          fundamentalTraders[i].setCash(cash - (nextPrice*(Math.min((cash/nextPrice),(tmpDiff*tmpA)/nextPrice)*fracFilledBuyOrders)));
          fundamentalTraders[i].setBuying(true);
        }
      } // END Else the trader is not in a fundamental cycle
    } // END For Rationing for Fundamental Traders
    // Rationing for the FUNDAMENTAL-TECHNICAL TRADERS
    for (int i = 0; i < numberTechnicalTraders; i++ ){
      stocks = technicalTraders[i].getStocks();
      cash = technicalTraders[i].getCash();
      rate = technicalTraders[i].getRate();
      days = technicalTraders[i].getDays();
      LimitOrderBook tempLimitOrderRec;
      numPendingSellsOrBuys = technicalTraders[i].getLimitOrderBookSize();
      //  Ask if the trader has no limit orders to be reviewed
      if (numPendingSellsOrBuys == 0){
        if (technicalTraders[i].getFundamentalBehaviour()){
          if (technicalTraders[i].isSelling() && (stocks>0)){
            technicalTraders[i].setNumberShares( stocks - ((Math.min(stocks, (tmpDiff*tmpA)/nextPrice))*fracFilledSellOrders) );
            technicalTraders[i].setCash(cash + (nextPrice*((Math.min(stocks,(tmpDiff*tmpA)/nextPrice))*fracFilledSellOrders)) );
            if (technicalTraders[i].exitPosition(nextPrice,fundamentalValue)){
              technicalTraders[i].setDecision(HOLD);
            }
          }
          else if (technicalTraders[i].isBuying() && (cash>0)){
            technicalTraders[i].setNumberShares( stocks + (Math.min((cash/nextPrice),(tmpDiff*tmpA)/nextPrice)*fracFilledBuyOrders) );
            technicalTraders[i].setCash(cash - (nextPrice*(Math.min((cash/nextPrice),(tmpDiff*tmpA)/nextPrice)*fracFilledBuyOrders)));
            if (technicalTraders[i].exitPosition(nextPrice,fundamentalValue)){
              technicalTraders[i].setDecision(HOLD);
            }
          }
          else{
            // Review the fundamental process
            tempDes = technicalTraders[i].getDecision();
            if (tempDes==SELL && (stocks>0)){
              technicalTraders[i].setNumberShares( stocks - ((Math.min(stocks, (tmpDiff*tmpA)/nextPrice))*fracFilledSellOrders) );
              technicalTraders[i].setCash(cash + (nextPrice*((Math.min(stocks,(tmpDiff*tmpA)/nextPrice))*fracFilledSellOrders)) );
              technicalTraders[i].setSelling(true);
            }
            else if(tempDes==BUY && (cash>0)){
              technicalTraders[i].setNumberShares( stocks + (Math.min((cash/nextPrice),(tmpDiff*tmpA)/nextPrice)*fracFilledBuyOrders) );
              technicalTraders[i].setCash(cash - (nextPrice*(Math.min((cash/nextPrice),(tmpDiff*tmpA)/nextPrice)*fracFilledBuyOrders)));
              technicalTraders[i].setBuying(true);
            }
            else{
              // Makes a prediction
              tempDes = technicalTraders[i].getBestProgram().evaluation(tempShort, tradingData[index-1]);
              technicalTraders[i].setDecision(tempDes);
              // The current trader WANTS to SELL
              if(tempDes == SELL){
                // The current trader has shares and there are traders willing to buy
                if ((stocks > 0)&&(totalBids > 0)){
                  technicalTraders[i].setNumberShares( stocks - ((stocks*tradingProportion)*fracFilledSellOrders) );
                  technicalTraders[i].setCash(cash + (nextPrice*((stocks*tradingProportion)*fracFilledSellOrders)));
                  if (groupsParamRec[util.getGroupNumber(groupsParamRec,i)].getLimitOrderBuyGeneration()==GENERATE_LIMIT_ORDER_BUY){
                    //  Generation of limit orders to buy again later at a lower price
                    tempLimitOrderRec=new LimitOrderBook(index-64,nextPrice*(1.0-(rate/100.0)),BUY,(stocks*tradingProportion)*fracFilledSellOrders);
                    technicalTraders[i].addLimitOrderBook(tempLimitOrderRec);
                  }
                }
              } // END If the trader wants to sell
              // The current trader WANTS to BUY
              else if (tempDes == BUY){
                // The current trader has cash and there are traderes willing to sell
                if ((cash > 0)&&(totalOffers > 0)){
                  technicalTraders[i].setNumberShares( stocks + ((tradingProportion*(cash/nextPrice))*fracFilledBuyOrders) );
                  technicalTraders[i].setCash(cash-(((tradingProportion*(cash/nextPrice))*fracFilledBuyOrders)*nextPrice));
                  if (groupsParamRec[util.getGroupNumber(groupsParamRec,i)].getLimitOrderSellGeneration()==GENERATE_LIMIT_ORDER_SELL){
                    //  Take Profit Limit order
                    tempLimitOrderRec=new LimitOrderBook(index-64,nextPrice*(1.0+(rate/100.0)),SELL,(tradingProportion*(cash/nextPrice))*fracFilledBuyOrders);
                    technicalTraders[i].addLimitOrderBook(tempLimitOrderRec);
                  }
                }
              } // END Else if the trader wants to buy
            } // END Else the trader made a technical decision
          } // END Else the trader is not in a fundamental cycle
        }
        else {
          // Makes a prediction
          tempDes = technicalTraders[i].getBestProgram().evaluation(tempShort, tradingData[index-1]);
          technicalTraders[i].setDecision(tempDes);
          // The current trader WANTS to SELL
          if(tempDes == SELL){
            // The current trader has shares and there are traders willing to buy
            if ((stocks > 0)&&(totalBids > 0)){
              technicalTraders[i].setNumberShares( stocks - ((stocks*tradingProportion)*fracFilledSellOrders) );
              technicalTraders[i].setCash(cash + (nextPrice*((stocks*tradingProportion)*fracFilledSellOrders)));
              if (groupsParamRec[util.getGroupNumber(groupsParamRec,i)].getLimitOrderBuyGeneration()==GENERATE_LIMIT_ORDER_BUY){
                //  Generation of limit orders to buy again later at a lower price
                tempLimitOrderRec=new LimitOrderBook(index-64,nextPrice*(1.0-(rate/100.0)),BUY,(stocks*tradingProportion)*fracFilledSellOrders);
                technicalTraders[i].addLimitOrderBook(tempLimitOrderRec);
              }
            }
          } // END If the trader wants to sell
          // The current trader WANTS to BUY
          else if (tempDes == BUY){
            // The current trader has cash and there are traderes willing to sell
            if ((cash > 0)&&(totalOffers > 0)){
              technicalTraders[i].setNumberShares( stocks + ((tradingProportion*(cash/nextPrice))*fracFilledBuyOrders) );
              technicalTraders[i].setCash(cash-(((tradingProportion*(cash/nextPrice))*fracFilledBuyOrders)*nextPrice));
              if (groupsParamRec[util.getGroupNumber(groupsParamRec,i)].getLimitOrderSellGeneration()==GENERATE_LIMIT_ORDER_SELL){
                //  Take Profit Limit order
                tempLimitOrderRec=new LimitOrderBook(index-64,nextPrice*(1.0+(rate/100.0)),SELL,(tradingProportion*(cash/nextPrice))*fracFilledBuyOrders);
                technicalTraders[i].addLimitOrderBook(tempLimitOrderRec);
              }
            }
          } // END Else if the trader wants to buy
        } // END Else the trader has no fundamental mechanism
      } // END If the tech trader has no limit order
      //  The current trader HAS a limit order to SELL
      else if (validLimitOrder(i,index-1,SELL)){
        technicalTraders[i].setDecision((byte)SELL);
        tempLimitOrderRec = new LimitOrderBook();
        tempLimitOrderRec = technicalTraders[i].getLimitOrder(0);
        double limitOrderSells = tempLimitOrderRec.getNumShares();
        time = tempLimitOrderRec.getTime();
        opType = tempLimitOrderRec.getOpType();
        date = (new Integer(tradingData[index-1].getDate())).intValue();
        //  There is someone willing to buy and the trader has shares
        if ((totalBids>0)&&(stocks>0)){
            technicalTraders[i].setNumberShares(stocks-(Math.min(limitOrderSells,stocks)*fracFilledSellOrders));
            technicalTraders[i].setCash(cash+(nextPrice*(Math.min(limitOrderSells,stocks)*fracFilledSellOrders)));
            execLimitOrder(i,index-1,SELL);
        }
        else if ((date-time) >= days){
          execLimitOrder(i,index-1,SELL);
        }
      } // END IF Limit order to sell
      //  The current trader HAS a limit order to BUY
      else if (validLimitOrder(i,index-1,BUY)){
        tempLimitOrderRec = new LimitOrderBook();
        tempLimitOrderRec = technicalTraders[i].getLimitOrder(0);
        double limitOrderBuys = tempLimitOrderRec.getNumShares();
        time = tempLimitOrderRec.getTime();
        opType = tempLimitOrderRec.getOpType();
        date = (new Integer(tradingData[index-1].getDate())).intValue();
        // There is someone willing to sell and the trader has cash
        if ((totalOffers>0)&&(cash>0)){
          if  (cash >= ((limitOrderBuys*nextPrice)*fracFilledBuyOrders)){
            technicalTraders[i].setNumberShares(stocks+(limitOrderBuys*fracFilledBuyOrders));
            technicalTraders[i].setCash(cash-((limitOrderBuys*nextPrice)*fracFilledBuyOrders));
            execLimitOrder(i,index-1,BUY);
          }
          else{
            technicalTraders[i].setNumberShares(stocks+((cash/nextPrice)*fracFilledBuyOrders));
            technicalTraders[i].setCash( cash - (((cash/nextPrice)*fracFilledBuyOrders)*nextPrice) );
            execLimitOrder(i,index-1,BUY);
          }
        }
        else if ((date-time) >= days){
          execLimitOrder(i,index-1,BUY);
        }
      }// END IF Valid limit order to buy
    } // END For all tech traders
  } // END method updateCash&Stocks
  // Price determination method
  private static double updatePrice(int index){
    double totalBids = tradingRecords[index].getBids();
    double totalOffers = tradingRecords[index].getOffers();
    double totalBuyers = tradingRecords[index].getNumBuyers();
    double totalSellers = tradingRecords[index].getNumSellers();
    double price, stocks, cash, wealth;
    if (totalBids==0 && totalOffers==0){
     price = tradingData[index-1].getObservation();
    }
    else if (marketParamRec.getPriceEquation() == CHEN){
      price = safeValue((tradingData[index-1].getObservation()*(1+betha(totalBids, totalOffers))));
    }
    else{
      price = safeValue(tradingData[index-1].getObservation()+((totalBids-totalOffers)/
      (((double)(totalBuyers+totalSellers))*marketParamRec.getMarketDepth())));
    }
    for (int i = 0; i < technicalTraders.length; i++ ){
      stocks = technicalTraders[i].getStocks();
      cash = technicalTraders[i].getCash();
      technicalTraders[i].setWealth(cash+(stocks*price));
      wealth = technicalTraders[i].getWealth();
    }
    for (int i = 0; i < noisyTraders.length; i++ ){
      noisyTraders[i].setWealth(noisyTraders[i].getCash()+(noisyTraders[i].getNumberShares() * price));
    }
    for (int i = 0; i < fundamentalTraders.length; i++ ){
      fundamentalTraders[i].setWealth(fundamentalTraders[i].getCash()+(fundamentalTraders[i].getNumberShares()*price));
    }
    parent_app.printStatus("Trading round: " + (index-Filter_63_Window) + "  # Bids: " + totalBids + "   # Offers: " + totalOffers +
                        "   Previous Price: " + tradingData[index-1].getObservation() + "   Actual Price: " + price);
    tradingRecords[index].setPrice(price);
    return price;
  }
  //  Hyperbolic tangent
  private static double tanh(double d){
    return ( (Math.exp(d)-Math.exp(-d))/(Math.exp(d)+Math.exp(-d)) );
  }
  //  Betha function used in Chen's price determination
  private static double betha(double buys, double sells){
    if (buys >= sells){
      return tanh(marketParamRec.getBetha1()*(buys-sells));
    }
    else {
      return tanh(marketParamRec.getBetha2()*(buys-sells));
    }
  }
  //  Method to secure no negative prices
  private static double safeValue(double d){
    if (d < 0.0){
      return 0.0;
    }
    return d;
  }
  //  Method that updates the fundamental value
  private static double updateFundamentalValue(int index){
    double temp=0;
    double mean = fundamentalParamRec.getMean(), stddev = fundamentalParamRec.getStandardDeviation();
    if (index == Filter_63_Window){
      temp = fundamentalParamRec.getInitialValue();
    }
    else{
      temp = tradingRecords[index-1].getFundamentalValue() + util.getRandomNormal(rand,mean,stddev);
    }
    return safeValue(temp);
  }
  ///////////////////////////////////////////////////////////////////////////////////
  //  Method that updates the wealth for each trader and detrmines the retrainning //
  ///////////////////////////////////////////////////////////////////////////////////
  private static void updateWealth(int index, FileOutputStream fileNameTT, FileOutputStream fileNameFT,FileOutputStream fileNameNT){
    double totalWealth = 0.0, totalCash = 0.0, totalShares = 0.0;
    int numRetrainTraders = 0;
    int memoryLenght = 0, retrainCond = FIXED, retrainPeriodicity = 0, cumulative = 0;
    for (int i = 0; i < technicalTraders.length; i++ ){
      totalCash+=technicalTraders[i].getCash();
      totalShares+=technicalTraders[i].getStocks();
      technicalTraders[i].setCumulativeCash(technicalTraders[i].getCumulativeCash()+technicalTraders[i].getCash());
      technicalTraders[i].setCumulativeStocks(technicalTraders[i].getCumulativeStocks()+technicalTraders[i].getStocks());
      totalWealth+=technicalTraders[i].getWealth();
    }
    for (int i = 0; i < noisyTraders.length; i++ ){
      totalCash+=noisyTraders[i].getCash();
      totalShares+=noisyTraders[i].getNumberShares();
      totalWealth+=noisyTraders[i].getWealth();
    }
    for (int i=0;i < fundamentalTraders.length; i++){
      totalCash+=fundamentalTraders[i].getCash();
      totalShares+=fundamentalTraders[i].getNumberShares();
      totalWealth+=fundamentalTraders[i].getWealth();
    }
    tradingRecords[index].setTotalWealth(totalWealth);
    tradingRecords[index].setTotalCash(totalCash);
    tradingRecords[index].setTotalShares(totalShares);
    tradingRecords[index].setAverageWealth(totalWealth/(technicalTraders.length+noisyTraders.length+fundamentalTraders.length));
    for (int i = 0; i < numberTechnicalTraders; i++ ){
      technicalTraders[i].setRetrain(0);
      cumulative = 0;
      if (technicalTraders[i].getLimitOrderBookSize()==0){
        int tmpGrNum = util.getGroupNumber(groupsParamRec,i);
        memoryLenght = groupsParamRec[tmpGrNum].getMemoryLenght();
        retrainCond = groupsParamRec[tmpGrNum].getRetrainningCondition();
        retrainPeriodicity = groupsParamRec[tmpGrNum].getFixedRetrainningPeriodicity();
        if (retrainCond == FIXED){
          //  Fixed Retraining Condition
          int tmp = (index+1/*+Filter_63_Window*/);
          if ( (tmp%retrainPeriodicity) == 0 ){
            DataRecord[] tempData;
            if ( index >= memoryLenght ){
              tempData = new DataRecord[memoryLenght];
              for (int j=0 ; j < memoryLenght ; j++ ){
                tempData[j] = tradingData[index-memoryLenght+j];
              }
            }
            else{
              tempData = new DataRecord[index];
              for (int j = 0; j < index ; j++ ){
                tempData[j] = tradingData[j];
              }
            }
            parent_app.printStatus("Fixed Training for trader :" + i);
            int[] tempRecommend = new int[tempData.length];
            tempRecommend = inputData.fillDecisions(technicalTraders[i].getRate(),technicalTraders[i].getDays(),tempData);
            technicalTraders[i].setRecommendations(tempRecommend);
            technicalTraders[i].train(tempData);
            technicalTraders[i].setRetrain(1);
            technicalTraders[i].setNumRetrains(technicalTraders[i].getNumRetrains()+1);
            numRetrainTraders++;
          }
        }
        //  Red Queen Retraining Condition
        else if ((technicalTraders[i].getWealth()<tradingRecords[index].getAverageWealth())){
          technicalTraders[i].setConsecutiveLowWealth(technicalTraders[i].getConsecutiveLowWealth()+1);
          if (technicalTraders[i].getConsecutiveLowWealth()==20){
            technicalTraders[i].setConsecutiveLowWealth(0);
            DataRecord[] tempData;
            if ( index >= memoryLenght ){
              tempData = new DataRecord[memoryLenght];
              for (int j=0 ; (j < memoryLenght) ; j++ ){
                tempData[j] = tradingData[index-memoryLenght+j];
              }
            }
            else{
              tempData = new DataRecord[index];
              for (int j = 0; j < index ; j++ ){
                tempData[j] = tradingData[j];
              }
            }
            parent_app.printStatus("Training under Red Queen for trader :" + i);
            int[] tempRecommend = new int[tempData.length];
            tempRecommend = inputData.fillDecisions(technicalTraders[i].getRate(),technicalTraders[i].getDays(),tempData);
            technicalTraders[i].setRecommendations(tempRecommend);
            technicalTraders[i].train(tempData);
            technicalTraders[i].setRetrain(1);
            technicalTraders[i].setNumRetrains(technicalTraders[i].getNumRetrains()+1);
            numRetrainTraders++;
          }
        }
        else{
          technicalTraders[i].setConsecutiveLowWealth(0);
        }
      }
    }
    tradingRecords[index].setnumRetrainTraders(numRetrainTraders);
    OutputData outputData = new OutputData();
    outputData.downLoadData(directoryName, technicalTraders, fileNameTT);
    outputData.downLoadData(directoryName, fundamentalTraders, fileNameFT);
    outputData.downLoadData(directoryName, noisyTraders, fileNameNT);
  }
  // Valid Limit Orders determination
  private static boolean validLimitOrder(int traderIndex, int recordIndex, int operType){
    boolean tempBool = false;
    LimitOrderBook tempLimitOrder = new LimitOrderBook();
    double observation = tradingData[recordIndex].getObservation();
    int days = technicalTraders[traderIndex].getDays();
    for (int i = 0; i < technicalTraders[traderIndex].getLimitOrderBookSize(); i++){
      tempLimitOrder = technicalTraders[traderIndex].getLimitOrder(i);
      double price = tempLimitOrder.getPrice();
      int time = tempLimitOrder.getTime();
      int opType = tempLimitOrder.getOpType();
      int date = (new Integer(tradingData[recordIndex].getDate())).intValue();
      if (((date-time)>=days) && (opType==operType)){
          tempBool = true;
      }
      else if (operType == SELL){
        if ((observation>price) && (opType==operType)){
          tempBool = true;
        }
      }
      else if (operType == BUY){
        if ((observation<price) && (opType==operType)){
          tempBool = true;
        }
      }
    }
    return tempBool;
  }
  // Execution of the limit order
  private static int execLimitOrder(int traderIndex, int recordIndex, int operType){
    int count = 0;
    LimitOrderBook tempLimitOrder = new LimitOrderBook();
    double observation = tradingData[recordIndex].getObservation();
    int days = technicalTraders[traderIndex].getDays();
    for (int i = 0; i < technicalTraders[traderIndex].getLimitOrderBookSize(); i++){
      tempLimitOrder = technicalTraders[traderIndex].getLimitOrder(i);
      double price = tempLimitOrder.getPrice();
      int time = tempLimitOrder.getTime();
      int opType = tempLimitOrder.getOpType();
      int date = (new Integer(tradingData[recordIndex].getDate())).intValue();
      if (((date-time)>=days) && (opType==operType)){
        technicalTraders[traderIndex].delLimitOrderBook(i);
        count ++;
      }
      else if (operType == SELL){
        if ( (observation > price) && (opType==operType) ){
         technicalTraders[traderIndex].delLimitOrderBook(i);
         count ++;
        }
      }
      else if (operType == BUY){
        if ( (observation < price) && (opType==operType) ){
         technicalTraders[traderIndex].delLimitOrderBook(i);
         count ++;
        }
      }
    }
    return count;
  }
  private static void saveData(String dirName, int epoch){
    OutputData outputData = new OutputData();
    outputData.downLoadData(dirName, tradingData, epoch+1);
    outputData.downLoadData(dirName, tradingRecords, epoch+1);
  }
  public void setParent(StockMarketApplet app)  {
    parent_app = app;
  }
  public int getTradingRound(){
    return this.tradingPeriod;
  }
  public int getMaxTradingRounds(){
    return this.marketParamRec.getMaxNumTradingPeriods();
  }
  public void runTest() {
    synchronized (runStatus) {
      if(runStatus.intValue() == IDLE) {
	runStatus = new Integer(RUN);
	market = new Thread(this);
	// This will cause a Security Exception in some browsers, but cause no harm
	market.setPriority(Thread.MIN_PRIORITY);
	market.start();
      }
    }
  }
  // Stop the Thread on the next possible opportunity.
  public void stopTest()  {
    synchronized (runStatus)    {
      if(runStatus.intValue() == RUN)
	runStatus = new Integer(STOP);
    }
  }
  //  The run() method is called when the Market object has created a new thread and invoked its start() method.
  //  This is the main loop of the market.
  public void run()  {
    int maxTradingPeriods;
//    this.init();
/*    if (!loaded){
      this.initialTraining();
    }*/
    this.parent_app.clear();
    maxTradingPeriods = marketParamRec.getMaxNumTradingPeriods();
    try{
      fileNameTT = new FileOutputStream(directoryName+"/OutputTechnicalTraders"+maxTradingPeriods+".csv");
      fileNameFT = new FileOutputStream(directoryName+"/OutputFundamentalTraders"+maxTradingPeriods+".csv");
      fileNameNT = new FileOutputStream(directoryName+"/OutputNoisyTraders"+maxTradingPeriods+".csv");
    }catch(FileNotFoundException e1){
      System.out.print("File Not Found Exception: " + e1.getMessage());
    }
    // Wait for the runtime environment to redraw all
    try { market.sleep(marketParamRec.getDelay()); }
    catch (InterruptedException irr)
      {}
    for (tradingPeriod=0 ; tradingPeriod < marketParamRec.getMaxNumTradingPeriods(); tradingPeriod++){
      tradingData[tradingPeriod + Filter_63_Window] = new DataRecord();
      // Starts evaluation of the current generation
      updateCashAndStocks(tradingPeriod + Filter_63_Window);
      tradingData[tradingPeriod + Filter_63_Window].setObservation(updatePrice(tradingPeriod + Filter_63_Window));
      computeIndicators(tradingPeriod + Filter_63_Window);
      updateWealth(tradingPeriod + Filter_63_Window, fileNameTT, fileNameFT, fileNameNT );
      sortedTechTraders=new TechnicalTrader[this.numberTechnicalTraders];
      for (int j=0 ; j<sortedTechTraders.length;j++ ){
        sortedTechTraders[j] = new TechnicalTrader();
        sortedTechTraders[j] = technicalTraders[j];
      }
      // Sort the population by their wealth
      util.QuickSort(sortedTechTraders,0,sortedTechTraders.length-1);
      // Display results in the applet
      parent_app.doneRun();
      // Wait for the graphic thread to draw
      try { market.sleep(marketParamRec.getDelay()); }
      catch (InterruptedException irr)
	{}
      if ((((tradingPeriod+1)%(marketParamRec.getSavingFrequency()))==0)&&(tradingPeriod!=0)){
        saveData(directoryName, tradingPeriod);
      }
      // Check if Thread shall be stopped
      synchronized (runStatus)
      {
	if(runStatus.intValue() == STOP)
	{
	  runStatus = new Integer(IDLE);
	  parent_app.stopped();
	  return;
	}
      }
    }
    try{
      fileNameTT.close();
      fileNameFT.close();
      fileNameNT.close();
    }catch(IOException e){
      System.out.print("IO Exception: " + e.getMessage());
    }
  }
  public double getPrice(){
    return tradingRecords[tradingPeriod + Filter_63_Window].getPrice();
  }
  public double getFundamentalValue(){
    return tradingRecords[tradingPeriod + Filter_63_Window].getFundamentalValue();
  }
  public TechnicalTrader getWorstTrader(){
    return sortedTechTraders[technicalTraders.length-1];
  }
  public TechnicalTrader getBestTrader(){
    return sortedTechTraders[0];
  }
  public TechnicalTrader getSelectedTrader(){
    return sortedTechTraders[selected];
  }
  public Program getSelectedCode(){
    return this.getSelectedTrader().getBestProgram();
  }
  public double getLastPrice(){
    return this.trainingData[trainingData.length-1].getObservation();
  }
  public double getBestWealth(){
    return this.getBestTrader().getWealth();
  }
  public double getWorstWealth(){
    return this.getWorstTrader().getWealth();
  }
  public double getAverageWealth(){
    return tradingRecords[tradingPeriod + Filter_63_Window].getAverageWealth();
  }
  public boolean checkTerminate()
  {
    return  (tradingPeriod >= marketParamRec.getMaxNumTradingPeriods()-1);
  }
  public int getBestSize(){
    return this.getBestTrader().getBestSize();
  }
  public int getWorstSize(){
    return this.getWorstTrader().getBestSize();
  }
  public int getAverageSize(){
    return this.getSelectedTrader().getAverageSize();
  }
  public int getPopulationSize(){
   return this.getSelectedTrader().getPopulationSize();
  }
  public int getSelectedDepth(){
    return this.getSelectedTrader().getBestProgram().depth();
  }
  public int getSelectedSize(){
    return this.getSelectedTrader().getBestSize();
  }
  public long getSelectedTime(){
    return this.getSelectedTrader().getSelectedTime();
  }
  public double getSelectedFitness(){
    return this.getSelectedTrader().getBestFitness();
  }
  public int getNumTechnicalTraders(){
    return this.numberTechnicalTraders;
  }
  protected void drawIndividual(int who, Individual one, Graphics g, Rectangle b)  {
    // If the computation thread is still running, DONT draw the current individual,
    synchronized (runStatus)
    {
      if(runStatus.intValue() != IDLE)
      {
	if(who == Market.DRAWIMP)
	  return;
      }
    }
    setDrawMode(true);
    setGraphics(g);
    setBounds(b);
    setDrawMode(false);
  }
  public void drawSelectedIndividual(int who, Graphics g, Rectangle bounds)
  {
//    drawIndividual(who, new Individual(technicalTraders[who].getSelectedProgram()), g, bounds);
  }
  // change the special individual using one of the navigation constants
  public void setSelected(int mode)
  {
    switch(mode)
    {
      case BEST:
	selected = 0;
	break;
      case WORST:
	selected = technicalTraders.length -1;
	break;
      case NEXT:
	if(selected < (technicalTraders.length -1))
	  selected++;
	break;
      case PREV:
	if(selected > 0)
	  selected--;
	break;
    }
  }
  private void setDrawMode(boolean mod)  {
    drawit= mod;
  }
  private void setGraphics(Graphics gr)  {
    graph= gr;
  }
  private void setBounds(Rectangle b)  {
    bounds = b;
  }

}