package stockMarketApplet.market.agents.technicalTrader;

import java.io.*;
import java.util.*;
import java.awt.Event;
import stockMarketApplet.util.Util;
import stockMarketApplet.market.io.*;
import stockMarketApplet.market.agents.technicalTrader.basics.*;

public class TechnicalTrader extends GEDEFIT
{
  private final byte BUY        = 1;
  private final byte SELL       = -1;
  private final byte HOLD       = 0;
  private final byte MV_12      = 10;
  private final int MV_50       = 11;
  private final int TRB_5       = 12;
  private final int TRB_50      = 13;
  private final int Filter_5    = 14;
  private final int Filter_63   = 15;
  private final int VOL_12      = 16;
  private final int VOL_50      = 17;
  private final int MOM_10      = 18;
  private final int MOM_60      = 19;
  private final int MMA_10      = 20;
  private final int MMA_60      = 21;
  private final byte PURE_TECH_BEHAVIOUR = 0;
  private final byte FUND_BEHAVIOUR = 1;
  private static byte DEPTH_CODE_GROWTH_CONTROL = 0;
  private static byte TARPEIAN_CODE_GROWTH_CONTROL = 1;
  private int n_tarpeian = 7;
  private double rate;
  private int days;
  private double wealth;
  private double stocks;
  private double cash;
  private double T;
  private double tau;
  private boolean selling;
  private boolean buying;
  private boolean fundamentalBehaviour;
  private int numRetraining;
  private int retrain;
  private int consecutiveLowWealth;
  private byte redQueenCounter;
  private double cumulativeRC;
  private double cumulativeRF;
  private double cumulativeRMC;
  private double cumulativeCash;
  private double cumulativeStocks;
  private double rc;
  private double rf;
  private double rmc;
  private byte decision;
  private int[] recommendations;
  private int[] forecasts;
  private Vector limitOrders;
  private FileOutputStream fileName;
  private FileOutputStream fileSummaryTraders;
  private FileInputStream fileNameInput;
  // Common constructor to other specific constructors
  public TechnicalTrader(){
    generator = new Random();
    Calendar date = Calendar.getInstance();
    generator.setSeed(date.get(Calendar.MILLISECOND));
    decisions = new ArrayList();
    decisions.add("BUY");
    decisions.add("SELL");
    decisions.add("HOLD");
    thresholds = new ArrayList();
    thresholds.add("0.0");
    limitOrders = new Vector();
    numRetraining = 0;
    consecutiveLowWealth = 0;
    redQueenCounter = 0;
    cumulativeCash = 0.0;
    cumulativeStocks = 0.0;
    decision = HOLD;
  }
  public TechnicalTrader(int maxMemTrainingData){
    recommendations = new int[maxMemTrainingData];
    forecasts =  new int[maxMemTrainingData];
  }

  // Constructor used to give a subset of indicators and functions used for mutation
  public TechnicalTrader(List ind, List logicFunct, List relFunct){
    this();
    indicators = new ArrayList();
    indicators = ind;
    logicFunctions = new ArrayList();
    logicFunctions.addAll(logicFunct);
    relationFunctions = new ArrayList();
    relationFunctions.addAll(relFunct);
    thresholds = new ArrayList();
    thresholds.add("0.0");
  }
  // Constructor used for valuations
  public TechnicalTrader(double rate, int days, int[] recommendations, double w_rc, double w_rf, double w_rmc){
    this();
    this.rate = rate;
    this.days = days;
    this.recommendations  = recommendations;
    this.forecasts = new int[recommendations.length];
    this.w_rc = w_rc;
    this.w_rf = w_rf;
    this.w_rmc = w_rmc;
  }
  // General fully parametrized constructor
  public TechnicalTrader(GroupParametersRecord groupsParamRec, GPParametersRecord gPParamRec, int maxMemTrainingData, int[] recommendations, List ind, List logicFuncts, List relFuncts){
    this();
    this.rate = groupsParamRec.getRate();
    this.days = groupsParamRec.getDays();
    this.cash = groupsParamRec.getCash();
    this.stocks = groupsParamRec.getNumberShares();
    if (groupsParamRec.getFundamentalBehaviour()==PURE_TECH_BEHAVIOUR){
      this.fundamentalBehaviour = false;
    }
    else{
      this.fundamentalBehaviour = true;
    }
    this.recommendations = new int[maxMemTrainingData];
    for (int i=0;i<recommendations.length;i++){
      this.recommendations[i]  = recommendations[i];
    }
    this.forecasts = new int[maxMemTrainingData];
    this.indicators = ind;
    this.logicFunctions = logicFuncts;
    this.relationFunctions = relFuncts;
    this.iniTreeDepth = gPParamRec.getIniTreeDepth();
    this.maxGens = gPParamRec.getMaxGenerations();
    this.maximizationStep = gPParamRec.getMaximizationStep();
    this.maxTreeDepth = gPParamRec.getMaxTreeDepth();
    this.minThreshold = gPParamRec.getMinThreshold();
    this.percentCopy = gPParamRec.getPercentCopy();
    this.populationSize = gPParamRec.getPopulationSize();
    this.probMaximization = gPParamRec.getProbMaximization();
    this.probMutation = gPParamRec.getProbMutation();
    this.rangeThreshold = gPParamRec.getRangeThreshold();
    this.selectionMethod = gPParamRec.getSelectionMethod();
    this.tournamentSize = gPParamRec.getTournamentSize();
    this.probSelectBest = gPParamRec.getProbSelectBest();
    this.codeGrowthControl = gPParamRec.getCodeGrowthControlMethod();
    this.n_tarpeian = gPParamRec.getNTarpeian();
    this.w_rc = gPParamRec.getWRC();
    this.w_rf = gPParamRec.getWRF();
    this.w_rmc = gPParamRec.getWRMC();
  }
  public void addLimitOrderBook(LimitOrderBook limitOrder){
    limitOrders.add(limitOrder);
  }
  public LimitOrderBook getLimitOrder(int index){
    return (LimitOrderBook)limitOrders.get(index);
  }
  public int getLimitOrderBookSize(){
    if (limitOrders.isEmpty())
      return 0;
    else
      return limitOrders.indexOf(limitOrders.lastElement()) + 1;
  }
  public void delLimitOrderBook(int index){
    limitOrders.removeElementAt(index);
  }
  protected void evalPopulation(DataRecord[] data)
  {
    Individual ind;
    generations[epoch].clear();
    generations[epoch].computeAvgSize();
    for(int p = 0; p < generations[epoch].getPopulationSize(); p++){
      if (this.codeGrowthControl == TARPEIAN_CODE_GROWTH_CONTROL){
        ind = evalInd(generations[epoch].getIndividual(p), data, generations[epoch].getAverageSize());
      }
      else{
        ind = evalInd(generations[epoch].getIndividual(p), data);
      }
      generations[epoch].getIndividual(p).setTime(ind.getTime());
      generations[epoch].getIndividual(p).setFitness(ind.getFitness());
    }
    generations[epoch].computeStats();
  }
  public Individual evalInd(Individual ind, DataRecord[] data, int avgSize)
  {
    long time = System.currentTimeMillis();
    //////////////////////////////////////////////////
    if ( (ind.size() > avgSize) && ((generator.nextInt()%this.n_tarpeian)==0) )
    {
      ind.setFitness((double) 0);
    }
    else{
      for(int i= 0; i < data.length/* - days*/; i++)
      {
        forecasts[i]=(ind.evaluation(data[i]));
      }
      ind.setFitness(fitnessComp(ind));
    }
    time = System.currentTimeMillis() - time;
    ind.setTime(time);
    return ind;
  }

  public Individual evalInd(Individual ind, DataRecord[] data)
  {
    long time = System.currentTimeMillis();
    //////////////////////////////////////////////////
    for(int i= 0; i < data.length/* - days*/; i++)
    {
      forecasts[i]=(ind.evaluation(data[i]));
    }
    ind.setFitness(fitnessComp(ind));
    time = System.currentTimeMillis() - time;
    ind.setTime(time);
    return ind;
  }

  protected double fitnessComp(Individual ind)
  {
    double rc = 0.0, rf = 0.0, rmc = 0.0;
    double fitnessFunction = 0.0, buysRatio = 0.0;
    int numTrueBuy = 0, numTrueSell = 0, numTrueHold = 0;
    int numFalseBuyActualSell = 0, numFalseBuyActualHold = 0;
    int numFalseSellActualBuy = 0, numFalseSellActualHold = 0;
    int numFalseHoldActualSell = 0, numFalseHoldActualBuy = 0;
    int tmpRec = 0,tmpFor = 0;
    for(int i= 0; i < recommendations.length - days; i++){
      tmpRec = recommendations[i];
      tmpFor = forecasts[i];
      if (tmpRec == BUY && tmpFor == BUY)
        numTrueBuy++;
      if (tmpRec == SELL && tmpFor == BUY)
        numFalseBuyActualSell++;
      if (tmpRec == HOLD && tmpFor == BUY)
        numFalseBuyActualHold++;
      if (tmpRec == SELL && tmpFor == SELL)
        numTrueSell++;
      if (tmpRec == BUY && tmpFor == SELL)
        numFalseSellActualBuy++;
      if (tmpRec == HOLD && tmpFor == SELL)
        numFalseSellActualHold++;
      if (tmpRec == HOLD && tmpFor == HOLD)
        numTrueHold++;
      if (tmpRec == SELL && tmpFor == HOLD)
        numFalseHoldActualSell++;
      if (tmpRec == BUY && tmpFor == HOLD)
        numFalseHoldActualBuy++;
    }
    rc = ((double)(numTrueBuy+numTrueSell+numTrueHold)/(double)(recommendations.length-days));
    rf = ((double)(numFalseBuyActualSell+numFalseBuyActualHold)/(double)(numFalseBuyActualSell+numTrueBuy+numFalseBuyActualHold));
    rmc = ((double)(numFalseSellActualBuy+numFalseHoldActualBuy)/(double)(numFalseSellActualBuy+numTrueBuy+numFalseHoldActualBuy));
    buysRatio = ((double)(numFalseBuyActualSell+numTrueBuy+numFalseBuyActualHold)/(double)(recommendations.length-days))*100.0;
    this.setRC(rc);
    this.setRF(rf);
    this.setRMC(rmc);
    fitnessFunction = (w_rc*rc) - (w_rf*rf) - (w_rmc*rmc);
//    return fitnessFunction;
    return rc;
  }
  public double getRate(){
    return this.rate;
  }
  public int getDays(){
    return this.days;
  }
  public double getStocks(){
    return this.stocks;
  }
  public double getCash(){
    return this.cash;
  }
  public double getWealth(){
    return this.wealth;
  }
  public double getT(){
    return this.T;
  }
  public double getTau(){
    return this.tau;
  }
  public boolean isSelling(){
    return this.selling;
  }
  public boolean isBuying(){
    return this.buying;
  }
  public double getDifference(){
    return (this.T-this.tau);
  }
  public boolean getFundamentalBehaviour(){
    return this.fundamentalBehaviour;
  }
  public int getNumRetrains(){
    return this.numRetraining;
  }
  public int getRetrain(){
    return this.retrain;
  }
  public int getConsecutiveLowWealth(){
    return this.consecutiveLowWealth;
  }
  public byte getRedQueenCounter(){
    return this.redQueenCounter;
  }
  public double getCumulativeRC(){
    return this.cumulativeRC;
  }
  public double getCumulativeRF(){
    return this.cumulativeRF;
  }
  public double getCumulativeRMC(){
    return this.cumulativeRMC;
  }
  public double getCumulativeCash(){
    return this.cumulativeCash;
  }
  public double getCumulativeStocks(){
    return this.cumulativeStocks;
  }
  public double getRC(){
    return this.rc;
  }
  public double getRF(){
    return this.rf;
  }
  public double getRMC(){
    return this.rmc;
  }
  public byte getDecision(){
    return this.decision;
  }
  public void setRate(double rate){
    this.rate = rate ;
  }
  public void setDays(int days){
    this.days = days;
  }
  public void setRecommendations(int[] serie){
    for (int i=0;i<serie.length;i++){
      this.recommendations[i] = serie[i];
    }
  }
  public void setNumberShares(double stocks){
    this.stocks = stocks;
  }
  public void setCash(double cash){
    this.cash = cash;
  }
  public void setWealth(double wealth){
    this.wealth = wealth;
  }
  public void setT(double t){
    this.T = t;
  }
  public void setTau(double tau){
    this.tau =  tau;
  }
  public void setFundamentalBehaviour(boolean fundBehaviour){
    this.fundamentalBehaviour = fundBehaviour;
  }
  public void setSelling(boolean selling){
    this.selling = selling;
  }
  public void setBuying(boolean buying){
    this.buying = buying;
  }
  public void setNumRetrains(int numRetrains){
    this.numRetraining = numRetrains;
  }
  public void setRetrain(int retrain){
    this.retrain = retrain;
  }
  public void setConsecutiveLowWealth(int consecutiveLowWealth){
    this.consecutiveLowWealth = consecutiveLowWealth;
  }
  public void setRedQueenCounter(byte redQueenCounter){
    this.redQueenCounter = redQueenCounter;
  }
  public void setCumulativeRC(double cumulativeRC){
    this.cumulativeRC = cumulativeRC;
  }
  public void setCumulativeRF(double cumulativeRFC){
    this.cumulativeRF = cumulativeRF;
  }
  public void setCumulativeRMC(double cumulativeRMC){
    this.cumulativeRMC = cumulativeRMC;
  }
  public void setCumulativeCash(double cumulativeCash){
    this.cumulativeCash = cumulativeCash;
  }
  public void setCumulativeStocks(double cumulativeStocks){
    this.cumulativeStocks = cumulativeStocks;
  }
  public void setRC(double rc){
    this.rc = rc;
  }
  public void setRF(double rf){
    this.rf = rf;
  }
  public void setRMC(double rmc){
    this.rmc = rmc;
  }
  public void setDecision(byte decision){
    this.decision = decision;
  }
  public byte takeDesicion(double price, double fundValue){
    byte decision=HOLD;
    if ((price-fundValue)>this.getT()){
      decision=SELL;
    }
    else if ((price-fundValue)<-this.getT()){
      decision=BUY;
    }
    return decision;
  }
  public boolean exitPosition(double price, double fundValue){
    boolean tempBool = false;
    if( (selling==true) && ((price-fundValue)<this.getTau()) ){
      selling = false;
      tempBool = true;
    }
    else if( (buying==true) && ((price-fundValue)>-this.getTau()) ){
      buying = false;
      tempBool = true;
    }
    return tempBool;
  }
  public void initialTraining(DataRecord[] data){
    makeFirstGen();
    do {
      if(generations[epoch].getValid()){
	makeNextGen(this.rate, this.days, data, this.recommendations);
      }
      evalPopulation(data);
    } while (!haveFinished());
  }
  public void copyLastGeneration(){
  }
  public void train(DataRecord[] data){
    epoch = 0;
    //  Copy the last generation from the previous training
    generations[epoch] = new Generation(generations[this.getMaxGenerations()].getPopulation());
    //////////////////////////////////////////////////////////////////////////////////////
    System.out.println("New population assigned ...");
    System.out.println();
    do {
      if(generations[epoch].getValid())
	makeNextGen(this.rate, this.days, data, this.recommendations);
      evalPopulation(data);
    } while (!haveFinished());
  }
  public void savePopulation(int groupNumber, int traderNumber){
//    Util util = new Util();
    String directoryName = new String();
    String lineBuffer = new String();
    String comma = new String(",");
    byte enter = Event.ENTER;
    String iString = new String();
    iString += groupNumber;
    iString += ".";
    iString += traderNumber;
    Calendar cal = new GregorianCalendar();
    int year = cal.get(Calendar.YEAR);             // 2002
    int month = cal.get(Calendar.MONTH)+1;         // 0=Jan, 1=Feb, ...
    int day = cal.get(Calendar.DAY_OF_MONTH);      // 1...
    int hour24 = cal.get(Calendar.HOUR_OF_DAY);     // 0..23
    int min = cal.get(Calendar.MINUTE);             // 0..59
    directoryName = ""+year+"-"+month+"-"+day+"."+hour24+min;
    directoryName +=  "TechnicalTraders/";
    boolean success = (new File(directoryName)).mkdir();
    if (!success) {
      // Directory creation failed
      System.out.println("Failure to create directory");
    }
    Util.copyParametersFiles(directoryName);
    //saving the population
    try{
      fileName = new FileOutputStream(directoryName + iString + ".txt");
    }catch(FileNotFoundException e1){
      System.out.print("File Not Found Exception: " + e1.getMessage());
    }
    try{
      lineBuffer = new String();
      for (int i=0;i<this.populationSize;i++){
        lineBuffer += this.generations[this.getMaxGenerations()].getPopulation()[i].toString();
        fileName.write(lineBuffer.getBytes());
        fileName.write('\r');
        fileName.write('\n');
        lineBuffer = new String();
      }
    } catch(IOException e) {
      System.out.print("IO Exception: " + e.getMessage());
    }
    try{
      fileName.close();
    }catch(IOException e){
      System.out.print("IO Exception: " + e.getMessage());
    }
    // saving the summary file
    try{
      fileSummaryTraders = new FileOutputStream(directoryName + "Summary"+ iString +".txt");
    }catch(FileNotFoundException e1){
      System.out.print("File Not Found Exception: " + e1.getMessage());
    }
    try{
      lineBuffer = new String();
      this.generations[this.getMaxGenerations()].computeAvgSize();
      lineBuffer = lineBuffer + "Average Size " + '\t' + this.generations[this.getMaxGenerations()].getAverageSize();
      fileSummaryTraders.write(lineBuffer.getBytes());
      fileSummaryTraders.write('\r');
      fileSummaryTraders.write('\n');
      lineBuffer = new String();

//      this.generations[this.getMaxGenerations()].computeAvgSize();
      lineBuffer = lineBuffer + "Average Fitness " + '\t' + this.generations[this.getMaxGenerations()].getAverageFitness();
      fileSummaryTraders.write(lineBuffer.getBytes());
      fileSummaryTraders.write('\r');
      fileSummaryTraders.write('\n');
      lineBuffer = new String();


    } catch(IOException e) {
      System.out.print("IO Exception: " + e.getMessage());
    }
    try{
      fileSummaryTraders.close();
    }catch(IOException e){
      System.out.print("IO Exception: " + e.getMessage());
    }



  }
  public void loadPopulation(int groupNumber, int traderNumber){
    List tmpLst=new ArrayList();
    String tmpStr = new String();
    Individual[] pop = new Individual[this.populationSize];
    byte enter = Event.ENTER;
    String iString = new String();
    iString += groupNumber;
    iString += ".";
    iString += traderNumber;
    try{
      fileNameInput = new FileInputStream("TechnicalTraders/"+iString+".txt");
    }catch(FileNotFoundException e1){
      System.out.print("File Not Found Exception: " + e1.getMessage());
    }
      tmpStr = new String();
      for (int i=0;i<this.populationSize;i++){
        tmpStr = readLine();
        tmpLst = Util.interpretString(tmpStr);
        double tmpDouble = new Double(tmpLst.get(0).toString()).doubleValue();
        tmpLst.remove(0);
        pop[i] = new Individual(new Program(tmpLst));
        pop[i].setFitness(tmpDouble);
      }
    this.generations = new Generation[this.getMaxGenerations()+1];
    this.generations[this.getMaxGenerations()] = new Generation(pop);
    this.epoch = this.getMaxGenerations();
    try{
      fileNameInput.close();
    }catch(IOException e){
      System.out.print("IO Exception: " + e.getMessage());
    }
  }
  private String readLine(){
    int c;
    int cr = 13;
    String bufferLine = new String();
    try{
        c = fileNameInput.read();
        while (c!=cr && c!=-1) {
          bufferLine += (char)(c);
          c = fileNameInput.read();
        }
        fileNameInput.skip(1);
    } catch(IOException e) {
      System.out.print("IO Exception: " + e.getMessage());
    }catch(ArrayIndexOutOfBoundsException e1){
      System.out.print("Array Index Out Of Bounds Exception: " + e1.getMessage());
    }
    return bufferLine;
  }
  public Object clone(){
    TechnicalTrader techTrader = new TechnicalTrader();
    techTrader.wealth = this.wealth;
    techTrader.generations = this.generations;
    return techTrader;
  }

}