package stockMarketApplet.market.agents.technicalTrader.basics;

import java.lang.Math;
import java.util.List;
import java.util.Random;
import java.util.Calendar;
import java.util.ArrayList;
import stockMarketApplet.market.io.DataRecord;

abstract public class GEDEFIT extends Object  {
  private static byte TOURNAMENT_SELECTION = 0;
  private static byte PROPORTIONATE_SELECTION = 1;
  private static byte DEPTH_CODE_GROWTH_CONTROL = 0;
  private static byte TARPEIAN_CODE_GROWTH_CONTROL = 1;
  /*  Table where the logical functions are stored  */
  protected  List logicFunctions;
  /*  Table where the relational functions are stored  */
  protected  List relationFunctions;
  /*  Table where the indicators are stored  */
  protected  List indicators;
  /*  Table where the decisions are stored  */
  protected  List decisions;
  /*  Table where the terminals are stored  */
  protected  List thresholds;
  protected  Random generator;
  protected  int populationSize;
  protected  Generation[] generations;
  protected  int maxGens;
  protected  int epoch;
  protected  double rangeThreshold;
  protected  double minThreshold;
  // Maximum tree depth allowed during crossover
  protected  byte maxTreeDepth;
  // Maximum tree depth at creation time
  protected  byte iniTreeDepth;
  // Percentage of parent to be identically copied
  protected  byte percentCopy;
  protected  byte probMutation;
  protected  byte probMaximization;
  protected  double maximizationStep;
  protected  byte selectionMethod;
  protected  int tournamentSize;
  protected  byte probSelectBest;
  protected double w_rc;
  protected double w_rf;
  protected double w_rmc;
  protected byte codeGrowthControl;
  protected int nTarpeian;
  public GEDEFIT(){
    generator = new Random();
    Calendar date = Calendar.getInstance();
    generator.setSeed(date.get(Calendar.MILLISECOND));
  }
  abstract protected void evalPopulation(DataRecord[] data);
  abstract public Individual evalInd(Individual ind, DataRecord[] data);
  abstract public Individual evalInd(Individual ind, DataRecord[] data, int popAvgSize);
  abstract protected double fitnessComp(Individual ind);
  protected  void clear()  {
    logicFunctions = null;
    relationFunctions = null;
    indicators = null;
    thresholds = null;
    decisions = null;
  }
  public  void setMaxGenerations(int i)  {
    maxGens= i;
  }
  public  int getMaxGenerations()  {
    return maxGens;
  }
  public  void setPopulationSize(int p)  {
      populationSize = p;
  }
  public  void setPercentCopy(byte n)  {
    if((n >= 0) && (n <= 100))
      percentCopy = n;
  }
  public  void setprobMutation(byte prob)  {
    if((prob >= 0) && (prob <= 100))
      probMutation = prob;
  }
  public  void setProbMaximization(byte prob)  {
    if((prob >= 0) && (prob <= 100))
      probMaximization = prob;
  }
  public  void setMaxTreeDepth(byte n)  {
    maxTreeDepth = n;
  }
  public  void setIniTreeDepth(byte n)  {
    iniTreeDepth = n;
  }
  public void setMinThreshold(double d){
    this.minThreshold = d;
  }
  public void setRangeThreshold(double d){
    this.rangeThreshold = d;
  }
  public  int getPercentCopy()  {
    return percentCopy;
  }
  public  byte getProbMutation()  {
    return probMutation;
  }
  public  byte getProbMaximization()  {
    return probMaximization;
  }
  public  int getEpoch()  {
    return epoch;
  }
  public  int getPopulationSize()  {
    return populationSize;
  }
  public  byte getMaxTreeDepth()  {
    return maxTreeDepth;
  }
  public  double getBestFitness()  {
    return generations[epoch].getBestIndividual().getFitness();
  }
  public  double getAverageFitness()  {
    return generations[epoch].getAverageFitness();
  }
  public  double getSelectedFitness()  {
    return generations[epoch].getSelected().getFitness();
  }
  public  int getBestSize()  {
    return generations[epoch].getBestIndividual().size();
  }
  public  int getAverageSize()  {
    return generations[epoch].getAverageSize();
  }
  public  int getSelectedSize()  {
    return generations[epoch].getSelected().size();
  }
  public  int getSelectedDepth()  {
    return generations[epoch].getSelected().depth();
  }
  public  long getSelectedTime()  {
    return generations[epoch].getSelected().getTime();
  }
  public  Program getSelectedProgram()  {
    return generations[epoch].getSelected().getProgram();
  }
  public  Program getBestProgram()  {
    return generations[epoch].getBestIndividual().getProgram();
  }
  public List getIndicators(){
    return indicators;
  }
  public List getDecisions(){
      return decisions;
  }
  public  void setSelected(int mode)  {
    generations[epoch].setSelected(mode);
  }
  public  Program createRoot(int depth)  {
    if(depth < 2)
    {
      System.out.println("Unable to create a root node of depth less than 2");
      return null;
    } else
    {
      List functite = new ArrayList();
      functite.add("ITE");
      functite.addAll(createRandCond(depth -1));
      functite.addAll(createRandTree(depth - 1));
      functite.addAll(createRandTree(depth - 1));
      return new Program(functite);
    }
  }
  public  List createRandCond(int depth)  {
    if(depth < 3)
    {
      return createRelFunct();
    } else
    {
      double p = generator.nextDouble();
      if( p <= 0.5){
        return createLogicFunct(depth);
      } else {
        return createRelFunct();
      }
    }
  }
  //  Selects randomly a relational function from the function array
  protected  List createRelFunct()  {
    int p = (int) (generator.nextInt(relationFunctions.size()));
    List functrel = new ArrayList();
    List tmpInd = new ArrayList();
    tmpInd.add("MOM_10");
    tmpInd.add("MOM_60");
    tmpInd.add("MMA_10");
    tmpInd.add("MMA_60");
    if (p == 0){
        functrel.add("<");
        String tmpStr = new String();
        tmpStr = createRandIndicator();
        functrel.add(tmpStr);
        if (tmpInd.contains(tmpStr)){
          functrel.add(createRandThreshold());
        }
        else{
          functrel.add(createRandNormalizedThreshold());
        }
        return functrel;
    }
    else if (p == 1){
        functrel.add(">");
        String tmpStr = new String();
        tmpStr = createRandIndicator();
        functrel.add(tmpStr);
        if (tmpInd.contains(tmpStr)){
          functrel.add(createRandThreshold());
        }
        else{
          functrel.add(createRandNormalizedThreshold());
        }
        return functrel;
    }
    else{
        functrel.add("=");
        String tmpStr = new String();
        tmpStr = createRandIndicator();
        functrel.add(tmpStr);
        if (tmpInd.contains(tmpStr)){
          functrel.add(createRandThreshold());
        }
        else{
          functrel.add(createRandNormalizedThreshold());
        }
        return functrel;
    }
  }
  public String createRandIndicator()  {
    return (String) indicators.get(  generator.nextInt( indicators.size() )  ) ;
  }
  public  String createRandThreshold()  {
    //  Let X uniform over [a,b]  we use a random Y uniform over [0,1]
    //  in the following way: X = a + (b-a) * Y
    double d = minThreshold + ( rangeThreshold * generator.nextDouble());
    return String.valueOf(d);
  }
  public  String createRandNormalizedThreshold()  {
    double d = -1+(2*generator.nextDouble());
    return String.valueOf(d);
  }
  //  Selects randomly a logical function from the function array*/
  protected  List createLogicFunct(int depth)  {
    int p = generator.nextInt(logicFunctions.size());
    List functlogic = new ArrayList();
    if (p == 0){
      functlogic.add("AND");
      functlogic.addAll(createRandCond(depth-1));
      functlogic.addAll(createRandCond(depth-1));
      return functlogic;
    }
    else if (p == 1){
      functlogic.add("OR");
      functlogic.addAll(createRandCond(depth-1));
      functlogic.addAll(createRandCond(depth-1));
      return functlogic;
    }
    else {
      functlogic.add("NOT");
      functlogic.addAll(createRandCond(depth-1));
      return functlogic;
    }
  }
  public  List createRandTree(int depth)  {
    List functITE = new ArrayList();
    if(depth < 3)
    {
      functITE.add(createRandDecision());
      return functITE;
    } else
    {
      double p = generator.nextDouble();
      if(p <= 0.5)
      {
        functITE.add("ITE");
        functITE.addAll(createRandCond(depth -1));
        functITE.addAll(createRandTree(depth - 1));
        functITE.addAll(createRandTree(depth - 1));
        return functITE;
      }
      else
      {
        functITE.add(createRandDecision());
        return functITE;
      }
    }
  }
  public  String createRandDecision()  {
    return (String) (decisions.get(generator.nextInt(decisions.size())));
  }
  // Method to create the initial population
  protected  void makeFirstGen()  {
    generations = new Generation[maxGens+1];
    epoch = 0;
    Individual[] population = new Individual[populationSize];
    for(int i = 0; i < populationSize; i++) {
      population[i] = new Individual(createRoot(iniTreeDepth));
    }
    generations[epoch] = new Generation(population);
  }
  public  boolean haveFinished()  {
    return (epoch >= maxGens);
  }
  // Method to make the next generation of individuals
  public  void makeNextGen(double r, int d, DataRecord[] data, int[] recommendations) {
    int i, l, best=0, competitor;
    double tempDouble, fbest;
    Individual[] population = new Individual[populationSize];
    Individual parents[], offspring[], mutated;
    //  Individuals to copy to the next generation by "Elitism"
    l = (population.length * percentCopy) / 100;
    for(i= 0; i < l; i++) {
      population[i]= (Individual) generations[epoch].getBestRandInd(i).clone();
    }
    //  Individuals to perform crossover
    while(l < population.length) {
      offspring = new Individual[2];
      parents = new Individual[2];
      //  Fitness proportionate selection
      if (this.selectionMethod==PROPORTIONATE_SELECTION){
        parents[1] = (Individual) generations[epoch].getFitPropInd().clone();
        parents[2] = (Individual) generations[epoch].getFitPropInd().clone();
      }
      //  Tournament selection
      else{
        for (i = 0;i < parents.length; i++){
          fbest = -1.0;
          for (byte j = 0; j < this.tournamentSize; j++ ){
            competitor = generator.nextInt(this.populationSize);
            if ((generations[epoch].getIndividual(competitor).getFitness()>fbest)&&(this.generator.nextFloat()<(this.probSelectBest/100.0))){
              fbest = generations[epoch].getIndividual(competitor).getFitness();
              best = competitor;
            }
          }
          parents[i] = generations[epoch].getIndividual(best);
        }
      }
      offspring = parents[0].crossover(parents[1]);
      for(i= 0; i < offspring.length; i++){
        if(offspring[i].depth() <= maxTreeDepth && offspring[i].depth() > 1) {
	  if(l < population.length) {
            offspring[i].resetFitness();
            population[l] = offspring[i];
  	    l++;
  	  }
  	}
      }
    }
    //  Individuals to perform mutation and maximization
    for( i = 0; i < population.length ; i++){

      tempDouble = generator.nextDouble();
      if (tempDouble <= (probMutation/100.0) ){
        mutated = (Individual) population[i].mutation( this.iniTreeDepth, this.minThreshold, this.rangeThreshold ,
                                          this.indicators, this.logicFunctions, this.relationFunctions).clone();
        if (mutated.depth() <= maxTreeDepth){
          population[i] = (Individual) mutated.clone();
        }
      }
      tempDouble = generator.nextDouble();
      if (tempDouble <= (probMaximization/100.0) ){
        population[i] = (Individual) population[i].maximization( (short)0,population[i].getProgram(), r, d, data, recommendations,
                                                    this.w_rc, this.w_rf,this.w_rmc).clone();
      }
    }
    generations[epoch].dropPopulation();
    epoch++;
    try {
      generations[epoch]= new Generation(population);
    }
    catch (ArrayIndexOutOfBoundsException e) { }
  }

  public void savePopulation(){
    this.generations[epoch].getPopulation();
  }

}