package stockMarketApplet.market.io;

import java.io.*;
import java.lang.Math;
import java.util.ArrayList;

public class InputData{

  private  final int BUY    = 1;
  private  final int SELL   = -1;
  private  final int HOLD   = 0;

  private int numberData;
  private String initialDate;
  private String finalDate;
  private FileInputStream fileName;

  public InputData(String initialDate, String finalDate, String fileNameString){
    this.initialDate = initialDate;
    this.finalDate = finalDate;
    try{
        fileName = new FileInputStream(fileNameString);
    } catch(FileNotFoundException e) {
        System.out.print("File Not Found");
    }
  }

  public int getNumberData(){
    return this.numberData;
  }

  public String getInitialDate(){
    return this.initialDate;
  }

  public String getFinalDate(){
    return this.finalDate;
  }

  public void setNumberData(int numberData){
    this.numberData = numberData;
  }

  public void setInitialDate(String initialDate){
    this.initialDate = initialDate;
  }

  public void setFinalDate(String finalDate){
    this.finalDate = finalDate;
  }

  public DataRecord[] loadData ( String fileNameString){
    final int dateCharacters = 10;
    byte[] bufferDate = new byte[dateCharacters];
    byte[] enter = new byte[2];
    boolean done = false;
    numberData  = this.getNumberRows();
    DataRecord[] dataArray = new DataRecord[numberData];
    try{
        fileName = new FileInputStream(fileNameString);
    } catch(FileNotFoundException e) {
        System.out.print("File Not Found");
    }
    try{
      while (!done ){
        fileName.read(bufferDate);
        if (new String(bufferDate).compareTo(initialDate) == 0)
        {
          done=true;
          dataArray[0] = new DataRecord();
          dataArray[0].setDate( new String(bufferDate) );
          dataArray[0].setObservation( (new Double(readNumber())).doubleValue() );
          dataArray[0].setMV_12( (new Double( readNumber() )).doubleValue() );
          dataArray[0].setMV_50( (new Double( readNumber() )).doubleValue() );
          dataArray[0].setTRB_5( (new Double( readNumber() )).doubleValue() );
          dataArray[0].setTRB_50( (new Double( readNumber() )).doubleValue() );
          dataArray[0].setFilter_5( (new Double( readNumber() )).doubleValue() );
          dataArray[0].setFilter_63( (new Double( readNumber() )).doubleValue() );
          dataArray[0].setVOL_12( (new Double( readNumber() )).doubleValue() );
          dataArray[0].setVOL_50( (new Double( readNumber() )).doubleValue() );
          dataArray[0].setMOM_10( (new Double( readNumber() )).doubleValue() );
          dataArray[0].setMOM_60( (new Double( readNumber() )).doubleValue() );
          dataArray[0].setMMA_10( (new Double( readNumber() )).doubleValue() );
          dataArray[0].setMMA_60( (new Double( readNumber() )).doubleValue() );
          fileName.read(enter);
        }
        else
        {
          readLine();
        }
      }
      for (int i=1 ; i < numberData ; i++)
      {
        dataArray[i] = new DataRecord();
        fileName.read(bufferDate);
        String tempString = new String(bufferDate);
        dataArray[i].setDate( new String(bufferDate) );
        dataArray[i].setObservation( (new Double(readNumber())).doubleValue() );
        dataArray[i].setMV_12( (new Double( readNumber() )).doubleValue() );
        dataArray[i].setMV_50( (new Double( readNumber() )).doubleValue() );
        dataArray[i].setTRB_5( (new Double( readNumber() )).doubleValue() );
        dataArray[i].setTRB_50( (new Double( readNumber() )).doubleValue() );
        dataArray[i].setFilter_5( (new Double( readNumber() )).doubleValue() );
        dataArray[i].setFilter_63( (new Double( readNumber() )).doubleValue() );
        dataArray[i].setVOL_12( (new Double( readNumber() )).doubleValue() );
        dataArray[i].setVOL_50( (new Double( readNumber() )).doubleValue() );
        dataArray[i].setMOM_10( (new Double( readNumber() )).doubleValue() );
        dataArray[i].setMOM_60( (new Double( readNumber() )).doubleValue() );
        dataArray[i].setMMA_10( (new Double( readNumber() )).doubleValue() );
        dataArray[i].setMMA_60( (new Double( readNumber() )).doubleValue() );
        fileName.read(enter);
      }
    } catch(IOException e) {
      System.out.print("IO Exception: " + e.getMessage());
    }catch(ArrayIndexOutOfBoundsException e1){
      System.out.print("Array Index Out Of Bounds Exception: " + e1.getMessage());
    }
    return dataArray;
  }

  public int[] fillDecisions(double rate, int days, DataRecord[] dataArray){
    numberData = dataArray.length;
    int[] temp = new int[numberData];
    for(int i = 0; i < numberData - days ; i++){
      temp[i] = (evaluateInvest(rate, days, dataArray, i));
    }
    return temp;
  }

  private int evaluateInvest (double rate, int days, DataRecord[] dataArray, int index){
    int decision = HOLD;
    int j = 0;
    boolean done = false;
    double dataTemp = 0.0;
    do {
      dataTemp = ( dataArray[index+j+1].getObservation()/dataArray[index].getObservation() )-1 ;
      if ( dataTemp > rate ){
        decision = BUY;
        done = true;
      } else if (dataTemp < -rate ){
        decision = SELL;
        done = true;
      }
      j++;
    }while ( (!done) && (j < days) );
    return decision;
  }

  private int getNumberRows(){
    int dateCharacters = 10;
    byte[] bufferDate = new byte[dateCharacters];
    String tempString;
    int temp = 0;
    boolean done = false;
    boolean eof = false;
    try
    {
      while (!done && !(eof)){
        fileName.read(bufferDate);
        if (new String (bufferDate).compareTo(initialDate) == 0){
          readLine();
          temp++;
          while (!done && !(eof)){
            fileName.read(bufferDate);
            tempString=new String(bufferDate);
            if (new String(bufferDate).compareTo(finalDate) == 0){

              done = true;
            }
            if (readLine().compareTo("-1") == 0)
            {
              eof = true;
            }
            temp++;
          }
        }
        else
        {
          readLine();
        }
      }
      fileName.close();
    }catch(IOException e) {
      System.out.print("IO Exception: " + e.getMessage());
    }catch(ArrayIndexOutOfBoundsException e1){
      System.out.print("Array Index Out Of Bounds Exception: " + e1.getMessage());
    }
    return temp;
  }

  private String readLine(){
    int c;
    int cr = 13;
    String bufferLine = new String();
    try{
        c = fileName.read();
        while (c!=cr && c!=-1) {
          bufferLine += (char)(c);
          c = fileName.read();
        }
        fileName.skip(1);
    } catch(IOException e) {
      System.out.print("IO Exception: " + e.getMessage());
    }catch(ArrayIndexOutOfBoundsException e1){
      System.out.print("Array Index Out Of Bounds Exception: " + e1.getMessage());
    }
    return bufferLine;
  }

  private String readNumber(){
    byte[] c = new byte[1];
    byte[] decimal = new byte[5];
    String bufferNumber = new String();
    try
    {
      do {
        fileName.read(c);
        bufferNumber += new String(c);
      } while (new String(c).compareTo(".") != 0);
      fileName.read(decimal);
      bufferNumber += new String(decimal);
    } catch(IOException e) {
      System.out.print("IO Exception: " + e.getMessage());
    }catch(ArrayIndexOutOfBoundsException e1){
      System.out.print("Array Index Out Of Bounds Exception: " + e1.getMessage());
    }
    return bufferNumber;
  }

}